/*
 * orientation.cpp
 *
 *  Created on: 26-05-2014
 *      Author: crm
 */

#include "orientation.h"

using namespace Engine;

//#define ORIENTATION_LEFT_HANDED

#ifndef ORIENTATION_LEFT_HANDED
#define CALC_UP AVectorNormalize(right^forward)
#define CALC_RIGHT AVectorNormalize(forward^up)
#define CALC_FORWARD AVectorNormalize(up^right)
#else
#define CALC_UP AVectorNormalize(forward^right)
#define CALC_RIGHT AVectorNormalize(up^forward)
#define CALC_FORWARD AVectorNormalize(right^up)
#endif

const float DEGTORAD=3.141592653589793/180.0;

const Orientation Orientation::FLAT_XY=Orientation(AVector(), AVector(0, 1, 0), AVector(0, 0, 1), AVector(1, 1, 1));
const Orientation Orientation::FLAT_XZ=Orientation(AVector(), AVector(0, 0, 1), AVector(0,-1, 0), AVector(1, 1, 1));
const Orientation Orientation::FLAT_YZ=Orientation(AVector(), AVector(0, 0, 1), AVector(1, 0, 0), AVector(1, 1, 1));
const Orientation Orientation::GUI=    Orientation(AVector(), AVector(0,-1, 0), AVector(0, 0,-1), AVector(1, 1, 1));

void Orientation::rotate(float deg, const AVector& axis)
	{
	AMatrix m=AMatrixRotate(deg, axis);

	forward=m*forward;
	right=m*right;
	up=CALC_UP;
	}

void Orientation::rotateRight(float deg)
	{
	float s, c;

	sincos(deg*DEGTORAD, s, c);

	forward=AVectorNormalize(forward*c+up*s);
	up=CALC_UP;
	}

void Orientation::rotateForward(float deg)
	{
	float s, c;

	sincos(deg*DEGTORAD, s, c);

	right=AVectorNormalize(right*c+up*s);
	up=CALC_UP;
	}

void Orientation::rotateUp(float deg)
	{
	float s, c;

	sincos(deg*DEGTORAD, s, c);

	right=AVectorNormalize(right*c+forward*s);
	forward=CALC_FORWARD;
	}

void Orientation::rotateX(float deg)
	{
	AMatrix m=AMatrixRotX(deg);

	right=m*right;
	forward=m*forward;
	up=CALC_UP;
	}

void Orientation::rotateY(float deg)
	{
	AMatrix m=AMatrixRotY(deg);

	right=m*right;
	forward=m*forward;
	up=CALC_UP;
	}

void Orientation::rotateZ(float deg)
	{
	AMatrix m=AMatrixRotZ(deg);

	right=m*right;
	forward=m*forward;
	up=CALC_UP;
	}


void Orientation::look(const AVector& target, const AVector& spos, const AVector& sup)
	{
	pos=spos;
	up=AVectorNormalize(sup);

	forward=AVectorNormalize(target-pos);
	right=CALC_RIGHT;
	up=CALC_UP;
	//right=CALC_RIGHT;
	}

void Orientation::look(const AVector& target, float distance, float elev, float deg)
	{
	float sa, ca;	// sin, cos; ang
	float se, ce;	// sin, cos; elev

	sincos(-deg*DEGTORAD, sa, ca);
	sincos(-elev*DEGTORAD, se, ce);

	forward=AVectorNormalize(AVector(sa*ce, ca*ce, se));
	//up=AVectorNormalize(AVector(sa*se, ca*se, ce));
	//right=CALC_RIGHT;
	right=AVector(ca, -sa, 0.0f);
	up=CALC_UP;

	pos=target-forward*distance;
	}


AMatrix Orientation::getMatrix() const
	{
	return AMatrixTranspose(AMatrix(right*scale.x, forward*scale.y, up*scale.z, AVector(0, 0, 0, 1)))*
	       AMatrix(AVector(1, 0, 0, pos.x), AVector(0, 1, 0, pos.y), AVector(0, 0, 1, pos.z), AVector(0, 0, 0, 1));
	}

AMatrix Orientation::getMatrixInverse() const
	{
	return AMatrix(AVector(1, 0, 0, -pos.x), AVector(0, 1, 0, -pos.y), AVector(0, 0, 1, -pos.z), AVector(0, 0, 0, 1))*
	       AMatrix(right*scale.x, forward*scale.y, up*scale.z, AVector(0, 0, 0, 1));
	}
