/*
 * Camera.cpp
 *
 *  Created on: 07-03-2013
 *      Author: crm
 */

#include "camera.h"

#include <GL/gl.h>
#include <cmath>

using namespace Engine;

void Camera::update()
	{
	//
	}

AMatrix Camera::getView() const
	{
	//return AMatrix(right*scale.x, up*scale.z, -forward*scale.y, AVector(pos.x, pos.y, pos.z, 1));
	//return getMatrixInverse();

	if(type==GUI)
		{
		return AMatrix(right*scale.x, up*scale.z, -forward*scale.y, AVector(0, 0, 0, 1));
		}
	else
		{
		return AMatrix(AVector(1, 0, 0, -pos.x), AVector(0, 1, 0, -pos.y), AVector(0, 0, 1, -pos.z), AVector(0, 0, 0, 1))*
			   AMatrix(right*scale.x, up*scale.z, -forward*scale.y, AVector(0, 0, 0, 1));
		}
	}

AMatrix Camera::getProjection() const
	{
	const float N=cutnear;
	const float F=cutfar;

	switch(type)
		{
		case ORTHOGRAPHIC:
			{
			const float TZ=-(F+N)/(F-N);

			return AMatrix(
					AVector(2.0f/width,           0,        0,  0),
					AVector(         0, 2.0f/height,        0,  0),
					AVector(         0,           0, -2/(F-N), TZ),
					AVector(         0,           0,        0,  1)
					);
			}
		break;

		case PERSPECTIVE:
			{
			const float ASPECT=width/height;
			const float FH=tan(fov/360.0*M_PI);
			const float FW=FH*ASPECT;
			const float C=-(F+N)/(F-N);
			const float D=-(2*F*N)/(F-N);

			return AMatrix(
					AVector(1/FW, 0, 0, 0),
					AVector(0, 1/FH, 0, 0),
					AVector(0, 0,  C, D),
					AVector(0, 0, -1, 0)
					);


			/*const float SW=2*N/width;
			const float SH=2*N/height;
			const float C=-(F+N)/(N-F);
			const float D=-(2*F*N)/(N-F);

			return AMatrix(
					AVector(SW,  0,  0, 0),
					AVector( 0, SH,  0, 0),
					AVector( 0,  0,  C, D),
					AVector( 0,  0, -1, 0)
					);*/
			}
		break;

		default:
		case GUI:
			{
			const float R=width;
			const float L=0;
			const float T=0;
			const float B=height;
			const float TX=-(R+L)/(R-L);
			const float TY=-(T+B)/(T-B);
			const float TZ=-(F+N)/(F-N);

			return AMatrix(
					AVector( 2/(R-L), 0, 0, TX),
					AVector(0,  2/(T-B), 0, TY),
					AVector(0, 0, -2/(F-N), TZ),
					AVector(0, 0, 0, 1)
					);
			}
		break;
		}

	return AMatrixIdentity(); // i tak tu nie dojdzie...
	}

AMatrix Camera::getProjection(int width, int height) const
	{
	const float N=cutnear;
	const float F=cutfar;

	switch(type)
		{
		case ORTHOGRAPHIC:
			{
			const float TZ=-(F+N)/(F-N);

			return AMatrix(
					AVector(2.0f/width,           0,        0,  0),
					AVector(         0, 2.0f/height,        0,  0),
					AVector(         0,           0, -2/(F-N), TZ),
					AVector(         0,           0,        0,  1)
					);
			}
		break;

		case PERSPECTIVE:
			{
			const float ASPECT=(float)width/height;
			const float FH=tan(fov/360.0*M_PI);
			const float FW=FH*ASPECT;
			const float C=-(F+N)/(F-N);
			const float D=-(2*F*N)/(F-N);

			return AMatrix(
					AVector(1/FW, 0, 0, 0),
					AVector(0, 1/FH, 0, 0),
					AVector(0, 0,  C, D),
					AVector(0, 0, -1, 0)
					);
			}
		break;

		default:
		case GUI:
			{
			const float R=width;
			const float L=0;
			const float T=0;
			const float B=height;
			const float TX=-(R+L)/(R-L);
			const float TY=-(T+B)/(T-B);
			const float TZ=-(F+N)/(F-N);

			return AMatrix(
					AVector( 2/(R-L), 0, 0, TX),
					AVector(0,  2/(T-B), 0, TY),
					AVector(0, 0, -2/(F-N), TZ),
					AVector(0, 0, 0, 1)
					);
			}
		break;
		}

	return AMatrixIdentity(); // i tak tu nie dojdzie...
	}


Orientation Camera::getBillboard(const AVector& objpos) const
	{
	switch(type)
		{
		case ORTHOGRAPHIC:
			{
			const AVector f=AVectorNormalize(AVector(forward.x, forward.y, 0.0f));
			const float ce=AVectorDot(forward, f).w;

			if(ce<0.1f && ce>-0.1f)
				return Orientation(objpos, AVector(0, 0, 1), -f, AVector(1, 0.0f, 1));
			return Orientation(objpos, AVector(0, 0, 1), -f, AVector(1, 1.0f/ce, 1));
			}
		break;

		case PERSPECTIVE:
			{
			const AVector dist=objpos-pos;
			const AVector f=AVectorNormalize(AVector(dist.x, dist.y, 0.0f));

			return Orientation(objpos, AVector(0, 0, 1), -f, AVector(1, 1, 1));
			}
		break;

		default:
		case GUI:
			return Orientation::GUI;
		break;
		}
	}

AVector Camera::getPointOnScreen(const AVector& objpos) const
	{
	AVector ret;

	ret=objpos-pos;
	ret=getView()*ret;
	ret=getProjection()*ret;
	if(type==PERSPECTIVE)
		ret=ret/AVector(ret.z, ret.z, 1, 1);
	ret.x=(ret.x+1)*width/2;
	ret.y=(ret.y+1)*height/2;
	ret.z=0;
	ret.w=0;

	return ret;
	}

void Camera::getRay(int mx, int my, AVector& raypos, AVector& raydir) const
	{
	float x;
	float y;

	if(type==PERSPECTIVE)
		{
		const float aspect=(float)width/height;
		float fw, fh;
		fh=tan(fov/360.0*M_PI);
		fw=fh*aspect;

		x=( mx/width-0.5f)*2.0f;
		y=(-my/height+0.5f)*2.0f;

		x=fw*x;
		y=fh*y;

		raypos=pos + forward*cutnear + right*x + up*y;
		raydir=AVectorNormalize(raypos-pos);
		}
	else
		{
		x=( mx/(float)width-0.5f)*width;
		y=(-my/(float)height+0.5f)*height;

		raypos=pos - forward*cutnear + right*x + up*y;
		raydir=forward;
		}
	}
