/*
 * imagemanager2.cpp
 *
 *  Created on: 26-02-2013
 *      Author: crm
 */

#include "imagemanager.h"

#include <map>
#include <vector>
#include <GL/glee.h>
#include <GL/gl.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>

#include "resource.h"
#include "log.h"

#include "image.h"

namespace Engine
	{
	namespace ImageManager
		{
		std::vector<Image*> images;
		std::map<std::string, Image *> named;
		}

	namespace ImageManager
		{
		//std::string getName(Image *img);
		}
	}

using namespace Engine;

/**** Zewnetrzne ****/
/*void ImageManager::flush()
	{
	for(auto img: images)
		{
		if(img->refs()>0)
			continue;

		Log::debug("ImageManager.flush: Kasowanie \"%s\"", getName(img).c_str());

		free(it->second);
		}
	}*/

void ImageManager::clear()
	{
	if(images.size()<1)
		return;

	Log::info("ImageManager.clear: Czyszczenie, %u obrazków", images.size());

	for(auto img: images)
		free(img, false);

	images.clear();
	named.clear();

	Log::ok("ImageManager.clear: Wyczyszczono");
	}


/*void ImageManager::reload()
	{
	GLint ctex;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &ctex);

	Log::info("ImageManager.reload: Przeladowywanie %u rysunkow", images.size());

	for(Image* img: images)
		{
		if(!img->srf)
			{
			Log::debug("ImageManager.reload: Brak surface");
			continue;
			}

		// Generowanie tekstur
		glGenTextures(1, &img->tex);

		int err;
		err=glGetError();

		if(!img->tex || err!=GL_NO_ERROR)
			{
			Log::fail("ImageManager.reload: Nie udalo sie dostac tekstury");
			continue;
			}

		glBindTexture(GL_TEXTURE_2D, img->tex);

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img->w, img->h, 0, GL_BGRA, GL_UNSIGNED_BYTE, img->srf->pixels);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

		err=glGetError();
		if(err!=GL_NO_ERROR)
			{
			Log::fail("ImageManager.reload: Nie udalo sie wygenerowac tekstury\n\terr: %04X (%d)", err, err);
			continue;
			}

		Log::debug("ImageManager.reload: Wczytano [GLid: %u]", img->tex);
		}

	glBindTexture(GL_TEXTURE_2D, ctex);
	}*/


ImagePtr ImageManager::get(const std::string& path)
	{
	Image *img;

	if(named.find(path)!=named.end())
		{
		return named[path];
		}

	Log::info("ImageManager.get: Wczytywanie \"%s\"", path.c_str());

	// Alokacja struktury
	try
		{
		img=new Image();
		}
	catch(std::bad_alloc &)
		{
		Log::fail("ImageManager.get: Nie udalo sie zajac miejsca na rysunek");
		return ImagePtr(nullptr);
		}

	// Wczytywanie grafiki
	char *resGfx=Resource::load(path);
	if(resGfx==NULL)
		{
		Log::fail("ImageManager.get: Nie udalo sie wczytac grafiki \"%s\"", path.c_str());
		return ImagePtr(nullptr);
		}

	// Dekodowanie
	SDL_Surface *temp;
	temp=IMG_Load_RW(SDL_RWFromMem(resGfx, Resource::getSize()), 1);    // Ustawianie gfx do sprite
	delete [] resGfx;

	if(temp==NULL)
		{
		Log::fail("ImageManager.get: Nie udalo sie zdekompresowac grafiki \"%s\"\n\terr: %s", path.c_str(), IMG_GetError());
		return ImagePtr(nullptr);
		}

	// Zmiana formatu
	SDL_Surface *surface=SDL_DisplayFormatAlpha(temp);
	SDL_FreeSurface(temp);
	if(surface==NULL)
		{
		Log::fail("ImageManager.get: Nie udalo sie przekonwertowac grafiki \"%s\"", path.c_str());
		return ImagePtr(nullptr);
		}

	img->w=surface->w;
	img->h=surface->h;
	/*img->srf=surface;*/

	// Generowanie tekstur
	glGenTextures(1, &img->tex);

	int err;
	err=glGetError();

	if(!img->tex || err!=GL_NO_ERROR)
		{
		Log::fail("ImageManager.get: Nie udalo sie dostac tekstury dla \"%s\"", path.c_str());
		return ImagePtr(nullptr);
		}

	//glBindTexture(GL_TEXTURE_2D, img.tex[0]);
	//Render::bind(img->tex[0]);
	// todo zbindowac teksture
	GLint ctex;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &ctex);

	glBindTexture(GL_TEXTURE_2D, img->tex);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img->w, img->h, 0, GL_BGRA, GL_UNSIGNED_BYTE, surface->pixels);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);	// Przy pomniejszaniu - linear
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// Przy powiekszaniu - nearest - chce PIKSELEEEAAAAA
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);		// Tylko jesli wspolrzedne sa wieksze od 1 -> profit!... chyba.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);		// Zobaczymy jak bedzie juz cos widac.

	err=glGetError();
	if(err!=GL_NO_ERROR)
		{
		Log::fail("ImageManager.get: Nie udalo sie wygenerowac tekstury dla \"%s\"\n\terr: %04X (%d)", path.c_str(), err, err);
		return ImagePtr(nullptr);
		}

	SDL_FreeSurface(surface);

	Log::debug("ImageManager.get: Wczytano \"%s\" [GLid: %u]", path.c_str(), img->tex);

	images.push_back(img);
	named[path]=img;

	glBindTexture(GL_TEXTURE_2D, ctex);

	return ImagePtr(img);
	}

ImagePtr ImageManager::get(SDL_Surface* surface)
	{
	Image *img;

	Log::info("ImageManager.get: Wczytywanie z surface");

	// Alokacja struktury
	try
		{
		img=new Image();
		}
	catch(std::bad_alloc &)
		{
		Log::fail("ImageManager.get(surface): Nie udalo sie zajac miejsca na rysunek");
		return ImagePtr(nullptr);
		}

	img->w=surface->w;
	img->h=surface->h;
	/*img->srf=surface;*/

	// Generowanie tekstur
	glGenTextures(1, &img->tex);

	int err;
	err=glGetError();

	if(!img->tex || err!=GL_NO_ERROR)
		{
		Log::fail("ImageManager.get(surface): Nie udalo sie dostac tekstury");
		return ImagePtr(nullptr);
		}

	GLint ctex;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &ctex);

	glBindTexture(GL_TEXTURE_2D, img->tex);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img->w, img->h, 0, GL_BGRA, GL_UNSIGNED_BYTE, surface->pixels);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);	// Przy pomniejszaniu - linear
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// Przy powiekszaniu - nearest - chce PIKSELEEEAAAAA
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);		// Tylko jesli wspolrzedne sa wieksze od 1 -> profit!... chyba.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);		// Zobaczymy jak bedzie juz cos widac.

	err=glGetError();
	if(err!=GL_NO_ERROR)
		{
		Log::fail("ImageManager.get(surface): Nie udalo sie wygenerowac tekstury\n\terr: %04X (%d)", err, err);
		return ImagePtr(nullptr);
		}

	Log::debug("ImageManager.get(surface): Wczytano [GLid: %u]", img->tex);

	glBindTexture(GL_TEXTURE_2D, ctex);

	images.push_back(img);

	return ImagePtr(img);
	}

ImagePtr ImageManager::get(const char *data, int w, int h, Type type)
	{
	Log::debug("ImageManager.loadGfx(data): NOT IMPLEMENTED");
	return ImagePtr(nullptr);
	}

ImagePtr ImageManager::get(unsigned tex, int w, int h)
	{
	Image *img;

	Log::info("ImageManager.get: Tworzenie z tekstury");

	// Alokacja struktury
	try
		{
		img=new Image();
		}
	catch(std::bad_alloc &)
		{
		Log::fail("ImageManager.get(texture): Nie udalo sie zajac miejsca na rysunek");
		return ImagePtr(nullptr);
		}

	img->tex=tex;
	img->w=w;
	img->h=h;

	// Generowanie tekstur
	glGenTextures(1, &img->tex);

	int err;

	//glBindTexture(GL_TEXTURE_2D, img.tex[0]);
	//Render::bind(img->tex[0]);
	// todo zbindowac teksture
	GLint ctex;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &ctex);

	glBindTexture(GL_TEXTURE_2D, img->tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);	// Przy pomniejszaniu - linear
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// Przy powiekszaniu - nearest - chce PIKSELEEEAAAAA
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);		// Tylko jesli wspolrzedne sa wieksze od 1 -> profit!... chyba.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);		// Zobaczymy jak bedzie juz cos widac.

	err=glGetError();
	if(err!=GL_NO_ERROR)
		{
		Log::fail("ImageManager.get(texture): Nie udalo sie wygenerowac tekstury\n\terr: %04X (%d)", err, err);
		return ImagePtr(nullptr);
		}

	Log::debug("ImageManager.get(texture): Wczytano [GLid: %u]", img->tex);

	glBindTexture(GL_TEXTURE_2D, ctex);

	images.push_back(img);

	return ImagePtr(img);
	}


void ImageManager::free(const Image* img, bool erase)
	{
	if(img->refs()>0)
		{
		Log::debug("ImageManager.free: Rysunek ma %d referencji, nie jest kasowany", img->refs());
		return;
		}

	if(erase)
		{
		for(std::map<std::string, Image *>::iterator it=named.begin(); it!=named.end(); ++it)
			{
			if(it->second!=img)
				continue;

			named.erase(it);

			break;
			}

		for(std::vector<Image*>::iterator it=images.begin(); it!=images.end(); ++it)
			{
			if(*it!=img)
				continue;

			images.erase(it);

			break;
			}
		}

	if(img->tex)
		glDeleteTextures(1, &img->tex);

	Log::debug("ImageManager.free: Kasowanie [GLid: %u]", img->tex);

	delete img;
	}


