/*
 * render.cpp
 *
 *  Created on: 13 mar 2016
 *      Author: crm
 */

#include "render.h"

#include <cmath>
#include <map>
#include <gl/glee.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <SDL/SDL.h>

#include "log.h"
#include "resource.h"

#include "vertex.h"
#include "framebuffer.h"

#define OPENGL_ERROR() {GLuint err=glGetError(); if(err!=GL_NO_ERROR) {printf("ERROR[%d]:\n%s", __LINE__, gluErrorString(err));}}

using namespace Engine;

namespace Engine
	{
	namespace Render
		{
		struct State
			{
			State(): texture(0u), camera(nullptr), color(1, 1, 1, 1), shader(nullptr), fbo(nullptr) {}

			GLuint texture;

			Camera *camera;
			AVector color;

			Shader *shader;
			FrameBuffer *fbo;
			};
		}

	// Zmuienne
	namespace Render
		{
		SDL_Surface *screen=nullptr;
		bool fullscreen=false;

		Camera cDefault;
		Camera cGui;

		GLuint vboid=0u;

		std::vector<State> states;

		Shader *postprocess=nullptr;

		Shader sDefault;
		Shader sPrimitive;

		FrameBuffer fDefault;
		FrameBuffer fGui;
		}

	// Funkcje pomocnicze
	namespace Render
		{
		//
		}
	}

bool Render::init(int w, int h, const std::string& title, const std::string& icon, bool fullscreen)
	{
	Log::debug("Render.init: Startowanie...");

	if(title.length()>0)
		SDL_WM_SetCaption(title.c_str(), NULL);
	else
		SDL_WM_SetCaption("Title", NULL);

	if(icon.length()>0)
		{
		char *data=Resource::load(icon.c_str());
		SDL_RWops *rwops=SDL_RWFromMem(data, Resource::getSize());
		SDL_Surface *img=SDL_LoadBMP_RW(rwops, 1);
		if(img)
			SDL_WM_SetIcon(img, NULL);	// todo zwalnianie surface ikony	// meeeeeh
		if(data)
			delete [] data;
		}

	SDL_ShowCursor(0);

	if(!setResolution(w, h, fullscreen))
		{
		Log::fail("Render.init: Nie udalo sie ustawic trybu wideo");
		return false;
		}

	Log::info("Render.init: Vendor: %s", glGetString(GL_VENDOR));
	Log::info("Render.init: Renderer: %s", glGetString(GL_RENDERER));
	Log::info("Render.init: Version: %s", glGetString(GL_VERSION));
	Log::info("Render.init: Shader version: %s", glGetString(GL_SHADING_LANGUAGE_VERSION));

	//glClearColor(1.0f, 1.0f, 1.0f, 0.0f);
	glClearColor(1, 1, 1, 0);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	/*glEnable(GL_CULL_FACE);
	glFrontFace(GL_CW);
	glCullFace(GL_BACK);*/

	glEnable(GL_TEXTURE_2D);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.0f);

	//glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);

	/*glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();*/

	int err=glGetError();
	if(err!=GL_NO_ERROR)
		{
		Log::fail("Render.init: Nie udalo sie wystartowac OpenGLa (%d)", err);
		return false;
		}

	// Inicjowanie VBO do printow
	glGenBuffers(1, &vboid);

	if(!vboid)
		{
		Log::fail("Render.init: Nie udalo sie wystartowac OpenGLa (vbo)");
		return false;
		}

	//SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 0);

	// Zerowanie zmiennych
	cDefault=Camera(Camera::PERSPECTIVE, screen->w, screen->h, 1.0f, 10000.0f);
	cGui=Camera(Camera::GUI, screen->w, screen->h, -1.0f, 1.0f);
	//cameras[CAMERA_GUI].look(AVector(screen->w/2, screen->h/2, 0), AVector(screen->w/2, screen->h/2, 0), AVector(0, 1, 0));

	sDefault.load("shader/default");
	sPrimitive.load("shader/primitive");

	states.clear();
	states.push_back(State());

	setMode(MODE_DEFAULT);
	setCamera(&cDefault);
	setShader(sDefault);
	setFrameBuffer();

	return true;
	}

void Render::update()
	{
	setColor(AVector(1, 1, 1, 1));

	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0u);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	State &state=states.back();

	const AMatrix MVP=AMatrixIdentity();
	const Vertex data[]={
		Vertex(-1, -1, 0,   0, 0),
		Vertex(-1,  1, 0,   0, 1),
		Vertex( 1,  1, 0,   1, 1),
		Vertex( 1, -1, 0,   1, 0)
		};

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	//glDisable(GL_DEPTH_TEST);
	if(postprocess)
		{
		setShader(*postprocess);
		postprocess->setUniforms();

		glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
		glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);
		glUniform1i(state.shader->getUniform("time"), SDL_GetTicks());

		bind(fDefault.getTexture());
		glDrawArrays(GL_QUADS, 0, 4);
		}
	else
		{
		glBindFramebuffer(GL_READ_FRAMEBUFFER, fDefault.getFBO());
		glBlitFramebuffer(0, 0, screen->w, screen->h, 0, 0, screen->w, screen->h,  GL_COLOR_BUFFER_BIT, GL_NEAREST); // <- blitowanie depth buffera powoduje INVALID_OPERATION
		glBindFramebuffer(GL_READ_FRAMEBUFFER, states.front().fbo->getFBO());
		}

	setShader(sDefault);
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	bind(fGui.getTexture());
	glDrawArrays(GL_QUADS, 0, 4);

	//glEnable(GL_DEPTH_TEST);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

	SDL_GL_SwapBuffers();

	while(states.size()>1)
		states.pop_back();

	setColor(AVector(1, 1, 1, 1));
	setCamera(&cDefault);
	setShader(sDefault);
	setMode(MODE_DEFAULT);

	setFrameBuffer(fGui, true);
	setFrameBuffer(fDefault, true);
	}

void Render::clear()
	{
	if(screen)
		SDL_FreeSurface(screen);
	screen=nullptr;

	sDefault.clear();
	sPrimitive.clear();
	fDefault.clear();
	fGui.clear();
	states.clear();
	}


bool Render::setResolution(int w, int h, bool sfullscreen)
	{
	fullscreen=sfullscreen;

	screen=SDL_SetVideoMode(w, h, 32, SDL_OPENGL|(fullscreen?SDL_FULLSCREEN:0));

	if(!screen)
		return false;

	if(!fDefault.init(screen->w, screen->h))
		{
		Log::fail("Render.setResolution: Nie udalo sie zainicjowac FBO ekranu");
		return false;
		}
	if(!fGui.init(screen->w, screen->h))
		{
		Log::fail("Render.setResolution: Nie udalo sie zainicjowac FBO GUI");
		return false;
		}

	Log::ok("Render.setResolution: %dx%d@%dbpp", screen->w, screen->h, screen->format->BitsPerPixel);

	return true;
	}

bool Render::setFullscreen(bool fullscreen)
	{
	return setResolution(screen->w, screen->h, fullscreen);
	}

bool Render::toggleFullscreen()
	{
	return setFullscreen(!fullscreen);
	}


AVector Render::getScreenSize()
	{
	return AVector(screen->w, screen->h);
	}


void Render::pushState()
	{
	states.push_back(State());
	State &prev=states[states.size()-2];
	State &cur=states.back();

	cur.color=prev.color;
	cur.texture=prev.texture;
	cur.camera=prev.camera;
	cur.shader=prev.shader;
	cur.fbo=prev.fbo;
	}

void Render::popState()
	{
	if(states.size()<=1)
		return;

	//State &prev=states.back();
	State &cur=states[states.size()-2];

	bind(cur.texture);
	//setCamera(cur.camera); // nie potrzebne, kamera jest odczytywana w razie potrzeby
	setShader(*cur.shader);

	setFrameBuffer(*cur.fbo, cur.fbo!=&fDefault && cur.fbo!=&fGui);

	states.pop_back();
	}


void Render::bind(GLuint tid)
	{
	State &state=states.back();

	if(tid==state.texture)
		return;

	glBindTexture(GL_TEXTURE_2D, tid);

	state.texture=tid;
	}

void Render::bind(const ImagePtr& img)
	{
	State &state=states.back();

	if(img->tex==state.texture)
		return;

	glBindTexture(GL_TEXTURE_2D, img->tex);

	state.texture=img->tex;
	}

void Render::bind(const SpritePtr& spr)
	{
	bind(spr->getImage());
	}


void Render::setShader(Shader& shader)
	{
	State &state=states.back();

	if(&shader==state.shader)
		return;

	glUseProgram(shader.getProgramID());

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif

	state.shader=&shader;
	}

void Render::setEffect(Shader& sshader)
	{
	postprocess=&sshader;
	}

void Render::unsetEffect()
	{
	postprocess=nullptr;
	}

/*void Render::applyEffect()
	{
	setColor(AVector(1, 1, 1, 1));

	//setFrameBuffer();

	Vertex data[]={
		Vertex(-1, -1, 0,   0, 0),
		Vertex(-1,  1, 0,   0, 1),
		Vertex( 1,  1, 0,   1, 1),
		Vertex( 1, -1, 0,   1, 0)
		};

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	const AMatrix MVP=AMatrixIdentity();

	if(postprocess)
		{
		State &state=states.back();

		setShader(*postprocess);

		glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
		glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);
		glUniform1i(state.shader->getUniform("time"), SDL_GetTicks());
		}

	bind(fDefault.getTexture());

	glDisable(GL_DEPTH_TEST);
	glDrawArrays(GL_QUADS, 0, 4);
	glEnable(GL_DEPTH_TEST);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	}*/

void Render::setFrameBuffer(FrameBuffer& s, bool clear)
	{
	State &state=states.back();

	if(&s==state.fbo)
		return;

	//setCamera(&fDefault.getCamera());

	glBindFramebuffer(GL_FRAMEBUFFER, s.getFBO());
#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif

	glViewport(0, 0, s.getWidth(), s.getHeight());

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif

	if(clear)
		{
		glClearColor(1, 1, 1, 0);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif
		}

	state.fbo=&s;
	}

void Render::setFrameBuffer()
	{
	State &state=states.back();

	//setCamera(&fDefault.getCamera());

	glBindFramebuffer(GL_FRAMEBUFFER, fDefault.getFBO());
#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif

	glViewport(0, 0, fDefault.getWidth(), fDefault.getHeight());

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif

	state.fbo=&fDefault;
	}

/*void Render::setFrameBufferGUI()
	{
	State &state=states.back();

	//setCamera(&fDefault.getCamera());

	glBindFramebuffer(GL_FRAMEBUFFER, fGui.getFBO());
	glViewport(0, 0, fGui.getWidth(), fGui.getHeight());

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif

	state.fbo=&fGui;
	}*/

void Render::setMode(Mode mode)
	{
	switch(mode)
		{
		case MODE_GUI:
			setCamera(&cGui);
			setFrameBuffer(fGui, false);
			//glDisable(GL_DEPTH_TEST);

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif
		break;

		case MODE_GUI_ITEM:
			setCamera(&cGui);
			//setFrameBuffer(fGui, false);
			//glDisable(GL_DEPTH_TEST);

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif
		break;

		default:
		case MODE_DEFAULT:
			setCamera(&cDefault);
			setFrameBuffer(fDefault, false);

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif

			//glEnable(GL_DEPTH_TEST);

#ifdef BUILD_DEBUG
	OPENGL_ERROR();
#endif
		break;;
		}
	}

void Render::setCamera(Camera *cam)
	{
	State &state=states.back();

	state.camera=cam;

	//glViewport(0, 0, cam->getWidth(), cam->getHeight());
	}

void Render::setColor(const AVector& color)
	{
	State &state=states.back();

	state.color=color;
	}

void Render::setAlpha(float alpha)
	{
	State &state=states.back();

	state.color.w=alpha;
	}


void Render::drawLine(const AVector& a, const AVector& b)
	{
	Vertex data[]={
		Vertex(a.x, a.y, a.z,   0.0f, 0.0f),
		Vertex(b.x, b.y, b.z,   1.0f, 1.0f)
		};

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	pushState();
	setShader(sPrimitive);
	State &state=states.back();

	AMatrix MVP=AMatrixTranspose(state.camera->getView()*state.camera->getProjection(state.fbo->getWidth(), state.fbo->getHeight()));
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	glDisable(GL_TEXTURE_2D);
	glDrawArrays(GL_LINES, 0, 2); // TODO dla kazdego sprite wygenerowac VBO z kazda klatka (autoincrement dla klatek), wykorzystac mozliwosc wyrysowywania fragmentow VBO
	glEnable(GL_TEXTURE_2D);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

	popState();
	}

void Render::drawTriangle(const AVector& a, const AVector& b, const AVector& c)
	{
	Vertex data[]={
		Vertex(a.x, a.y, a.z,   0.0f, 0.0f),
		Vertex(b.x, b.y, b.z,   1.0f, 0.0f),
		Vertex(c.x, c.y, c.z,   0.0f, 1.0f)
		};

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	pushState();
	setShader(sPrimitive);
	State &state=states.back();

	AMatrix MVP=AMatrixTranspose(state.camera->getView()*state.camera->getProjection(state.fbo->getWidth(), state.fbo->getHeight()));
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	glDisable(GL_TEXTURE_2D);
	glDrawArrays(GL_TRIANGLES, 0, 3);
	glEnable(GL_TEXTURE_2D);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

	popState();
	}

void Render::drawQuad(const AVector& a, const AVector& b, const AVector& c, const AVector& d)
	{
	Vertex data[]={
		Vertex(a.x, a.y, a.z,   0.0f, 0.0f),
		Vertex(b.x, b.y, b.z,   1.0f, 0.0f),
		Vertex(c.x, c.y, c.z,   1.0f, 1.0f),
		Vertex(d.x, d.y, d.z,   0.0f, 1.0f)
		};

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	pushState();
	setShader(sPrimitive);
	State &state=states.back();

	AMatrix MVP=AMatrixTranspose(state.camera->getView()*state.camera->getProjection(state.fbo->getWidth(), state.fbo->getHeight()));
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	glDisable(GL_TEXTURE_2D);
	glDrawArrays(GL_QUADS, 0, 4);
	glEnable(GL_TEXTURE_2D);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

	popState();
	}

void Render::drawRectLine(const AVector& a, const AVector& b)
	{
	Vertex data[]={
		Vertex(a.x, a.y, a.z,   0.0f, 0.0f),
		Vertex(a.x, b.y, a.z,   1.0f, 0.0f),
		Vertex(b.x, b.y, a.z,   1.0f, 1.0f),
		Vertex(b.x, a.y, a.z,   0.0f, 1.0f),
		Vertex(a.x, a.y, a.z,   0.0f, 0.0f)
		};

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	pushState();
	setShader(sPrimitive);
	State &state=states.back();

	AMatrix MVP=AMatrixTranspose(state.camera->getView()*state.camera->getProjection(state.fbo->getWidth(), state.fbo->getHeight()));
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	glDisable(GL_TEXTURE_2D);
	//glLineWidth(2.0f);
	glDrawArrays(GL_LINE_STRIP, 0, 5);
	//glLineWidth(1.0f);
	glEnable(GL_TEXTURE_2D);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

	popState();
	}

void Render::drawTexturedTriangle(const AVector& a, const AVector& b, const AVector& c,
		const AVector& ta, const AVector& tb, const AVector& tc)
	{
	State &state=states.back();

	Vertex data[]={
		Vertex(a.x, a.y, a.z,   ta.x, ta.y),
		Vertex(b.x, b.y, b.z,   tb.x, tb.y),
		Vertex(c.x, c.y, c.z,   tc.x, tc.y)
		};

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	AMatrix MVP=AMatrixTranspose(state.camera->getView()*state.camera->getProjection(state.fbo->getWidth(), state.fbo->getHeight()));
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	glDrawArrays(GL_TRIANGLES, 0, 3);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	}

void Render::drawTexturedRect(const AVector& a, const AVector& b, const AVector& c, const AVector& d,
		const AVector& ta, const AVector& tb, const AVector& tc, const AVector& td)
	{
	State &state=states.back();

	Vertex data[]={
		Vertex(a.x, a.y, a.z,   ta.x, ta.y),
		Vertex(b.x, b.y, b.z,   tb.x, tb.y),
		Vertex(c.x, c.y, c.z,   tc.x, tc.y),
		Vertex(d.x, d.y, d.z,   td.x, td.y)
		};

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	AMatrix MVP=AMatrixTranspose(state.camera->getView()*state.camera->getProjection(state.fbo->getWidth(), state.fbo->getHeight()));
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	glDrawArrays(GL_QUADS, 0, 4);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	}


void Render::print(const FrameBuffer& fb, const Orientation& orientation, float x, float y, float w, float h)
	{
	State &state=states.back();

	const float iw=fb.getWidth();
	const float ih=fb.getHeight();

	const Vertex data[]=
		{
		Vertex(0, 0, 0,  x   /iw,  y   /ih),
		Vertex(w, 0, 0, (x+w)/iw,  y   /ih),
		Vertex(w,-h, 0, (x+w)/iw, (y+h)/ih),
		Vertex(0,-h, 0,  x   /iw, (y+h)/ih),
		};

	bind(fb.getTexture());

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	const AMatrix MVP=AMatrixTranspose(orientation.getMatrix()*state.camera->getView()*state.camera->getProjection(state.fbo->getWidth(), state.fbo->getHeight()));
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	glDrawArrays(GL_QUADS, 0, 4);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

	/*glBindFramebuffer(GL_READ_FRAMEBUFFER, fb.getFBO());
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, fbo.getFBO());
	glBlitFramebuffer(0, 0, fb.getWidth(), fb.getHeight(), 0, 0, screen->w/2, screen->h/2, GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT, GL_NEAREST);*/
	}

void Render::print(const Image& img, const Orientation& orientation, float x, float y, float w, float h)
	{
	State &state=states.back();

	const float iw=img.w;
	const float ih=img.h;

	Vertex data[]=
		{
		Vertex(0, 0, 0,  x   /iw,  y   /ih),
		Vertex(w, 0, 0, (x+w)/iw,  y   /ih),
		Vertex(w,-h, 0, (x+w)/iw, (y+h)/ih),
		Vertex(0,-h, 0,  x   /iw, (y+h)/ih),
		};

	bind(img.tex);

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	const AMatrix MVP=AMatrixTranspose(orientation.getMatrix()*state.camera->getView()*state.camera->getProjection(state.fbo->getWidth(), state.fbo->getHeight()));
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	glDrawArrays(GL_QUADS, 0, 4);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	}

void Render::print(const ImagePtr& iptr, const Orientation& orientation, float x, float y, float w, float h)
	{
	print(*iptr, orientation, x, y, w, h);
	}

void Render::print(const Sprite& spr, const Orientation& orientation, unsigned animation, unsigned direction, unsigned frame)
	{
	const Sprite::Animation& a=spr.getAnimation(animation);
	const Sprite::Direction& d=a.getDirection((Sprite::DirectionName)direction);
	const Sprite::Frame& f=d.getFrame(frame);

	const AVector pos=orientation.getPos()+orientation.getForward()*f.spoty-orientation.getRight()*f.spotx;

	print(spr.getImage(), Orientation(pos, orientation), f.x, f.y, f.w, f.h);
	}

void Render::print(const SpritePtr& sptr, const Orientation& orientation, bool rotate)
	{
	State &state=states.back();

	int direction=sptr.getDirectionID();

	// Zakladam, ze jest to billboard - inaczej moze wyjsc nieciekawie~
	if(rotate)
		{
		const AVector& f=AVectorNormalize(orientation.getPos()-state.camera->getPos());//camera->getForward();
		const AVector& forward=AVectorNormalize(AVector(f.x, f.y, 0.0f));
		const float fu=/*forward|AVector(0, -1, 0)*/ forward.y;
		const float fr=/*forward|AVector(1,  0, 0)*/ forward.x;
		const bool frgt=fr>0.0f;

		const float tmp=acosf(fu)*180.0f/M_PI;
		const float ang=((360-tmp)*frgt+(tmp)*!frgt);

		direction+=(ang/45.0f+0.5f);

		if(direction>=(int)Sprite::DirectionName::COUNT) direction-=(int)Sprite::DirectionName::COUNT;
		else if(direction<0)                             direction+=(int)Sprite::DirectionName::COUNT;
		}

	print(*sptr, orientation, sptr.getAnimationID(), direction, sptr.getFrame());
	}


void Render::skybox(const Image& img, float x, float y, float w, float h)
	{
	State &state=states.back();

	const float iw=img.w;
	const float ih=img.h;

	const Vertex data[]=
		{
		Vertex(-1,  1, 0,  x   /iw,  y   /ih),
		Vertex( 1,  1, 0, (x+w)/iw,  y   /ih),
		Vertex( 1, -1, 0, (x+w)/iw, (y+h)/ih),
		Vertex(-1, -1, 0,  x   /iw, (y+h)/ih),
		};

	bind(img.tex);

	glBindBuffer(GL_ARRAY_BUFFER, vboid);
	glBufferData(GL_ARRAY_BUFFER, sizeof(data), data, GL_STREAM_DRAW);

	glEnableVertexAttribArray(0);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, x));
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)offsetof(Vertex, tx));

	const AMatrix MVP=AMatrixIdentity();
	glUniformMatrix4fv(state.shader->getUniform("MVP"), 1, GL_FALSE, &MVP.row[0].x);
	glUniform4fv(state.shader->getUniform("uColor"), 1, &state.color.x);

	//glDisable(GL_DEPTH_TEST);
	glDrawArrays(GL_QUADS, 0, 4);
	//glEnable(GL_DEPTH_TEST);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	}

void Render::skybox(const ImagePtr& iptr, float x, float y, float w, float h)
	{
	skybox(*iptr, x, y, w, h);
	}

void Render::skybox(const Sprite& spr, unsigned animation, unsigned direction, unsigned frame)
	{
	const Sprite::Animation& a=spr.getAnimation(animation);
	const Sprite::Direction& d=a.getDirection((Sprite::DirectionName)direction);
	const Sprite::Frame& f=d.getFrame(frame);

	skybox(spr.getImage(), f.x, f.y, f.w, f.h);
	}

void Render::skybox(const SpritePtr& sptr)
	{
	skybox(*sptr, sptr.getAnimationID(), sptr.getDirectionID(), sptr.getFrame());
	}
