/*
 * sound2.cpp
 *
 *  Created on: 23-07-2013
 *      Author: crm
 */

#include <map>
#include <vector>
#include <assert.h>
#include <SDL/SDL_mixer.h>

#include "log.h"
#include "resource.h"

#include "sound.h"

#ifdef WIN32
	#define WINDOWS
#else
	#define LINUX
#endif

using namespace Engine;

bool Sound::init(int channels, int sampling)
	{
	if(!(Mix_Init(MIX_INIT_OGG)&MIX_INIT_OGG))
		{
		Log::fail("Sound.init: Nie udalo sie uruchomic dekodera ogg");
		return false;
		}

	if(Mix_OpenAudio(sampling, MIX_DEFAULT_FORMAT, 2, 2048)<0)
		{
		Log::fail("Sound.init: Nie udalo sie uruchomic dzwieku");
		return false;
		}

	if(Mix_AllocateChannels(channels)<channels)
		{
		Log::fail("Sound.init: Nie udalo sie zaalokowac kanalow (%d)", channels);
		return false;
		}

	/*{
		int n;
		n=Mix_GetNumChunkDecoders();
		Log::info("Sound.init: chunk decoder count: %d", n);
		for(int i=0; i<n; ++i)
			Log::info(" > %s", Mix_GetChunkDecoder(i));

		n=Mix_GetNumMusicDecoders();
		Log::info("Sound.init: music decoder count: %d", n);
		for(int i=0; i<n; ++i)
			Log::info(" > %s", Mix_GetChunkDecoder(i));
	}*/

	//streams.resize(channels);

	return true;
	}

void Sound::update(float /*dt*/)
	{
	/*for(std::vector<Stream>::iterator it=streams.begin(); it!=streams.end(); ++it)
		{
		Stream& s=*it;

		if(s.strm.get()==NULL || !s.strm->isPlaying())
			continue;

		float vol, pch, pan;

		vol=s.strm->getVolume();
		pch=s.strm->getPitchShift();
		pan=s.strm->getPan();
		}*/
	}

void Sound::clear()
	{
	for(std::vector<Sample *>::iterator it=samples.begin(); it!=samples.end(); ++it)
		delete *it;

	samples.clear();
	Mix_Quit();

	Log::ok("Sound.clear: Wyczyszczono");
	}


int Sound::sound(const std::string& name, float volume)
	{
	return play(name, -1, volume*volumesound, false, 0.0f, 1.0f);
	}

int Sound::music(const std::string& name, float volume)
	{
	return play(name, -1, volume*volumemusic,  true, 0.0f, 1.0f);
	}


void Sound::pause(int channel)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	//streams[channel].strm->stop();

	Mix_Pause(channel);
	}

void Sound::resume(int channel)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	//streams[channel].strm->play();
	Mix_Resume(channel);
	}

void Sound::stop(int channel)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	//streams[channel].strm->stop();
	//streams[channel].strm->reset();
	//Mix_Pause(channel);
	Mix_HaltChannel(channel);
	}

void Sound::rewind(int channel)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	//streams[channel].strm->reset();
	Mix_HaltChannel(channel);
	Mix_Resume(channel);
	// todo
	}


bool Sound::isPlaying(int channel)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	return Mix_Playing(channel);//streams[channel].strm->isPlaying();
	}


void Sound::setGlobalSoundVolume(float vol)
	{
	volumesound=vol;

	Mix_Volume(-1, vol);
	}

void Sound::setGlobalMusicVolume(float vol)
	{
	volumemusic=vol;

	Mix_Volume(-1, vol);
	}


void Sound::setPitch(int channel, float val, int /*timeout*/)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	//streams[channel].pitch=val;

	//streams[channel].strm->setPitchShift(val);
	}

void Sound::setVolume(int channel, float val, int /*timeout*/)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	//streams[channel].vol=val;

	//streams[channel].strm->setVolume(val);
	}

void Sound::setPan(int channel, float val, int /*timeout*/)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	//streams[channel].pan=val;

	//streams[channel].strm->setPan(val);
	}


float Sound::getGlobalSoundVolume()
	{
	return volumesound;
	}

float Sound::getGlobalMusicVolume()
	{
	return volumemusic;
	}


float Sound::getPitch(int channel)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	return 0;//streams[channel].strm->getPitchShift();
	}

float Sound::getVolume(int channel)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	return 0;//streams[channel].strm->getVolume();
	}

float Sound::getPan(int channel)
	{
	//assert(channel>=0 && channel<(int)streams.size());

	return 0;//streams[channel].strm->getPan();
	}


Sound::Sample* Sound::findSoundByName(const std::string& name)
	{
	for(std::vector<Sample *>::iterator it=samples.begin(); it!=samples.end(); ++it)
		{
		if(name.compare((*it)->name)==0)
			return *it;
		}

	return NULL;
	}


Sound::Sample* Sound::load(const std::string& name)
	{
	Sample *snd;

	snd=new Sample();

	snd->data=Resource::load(name);
	snd->size=Resource::getSize();

	if(!snd->data)
		{
		delete snd;

		return NULL;
		}

	snd->snd=Mix_LoadWAV_RW( SDL_RWFromMem(snd->data, snd->size), false);

	samples.push_back(snd);

	return snd;
	}

int Sound::play(const std::string& name, int channel, float volume, bool repeat, float pan, float pitch)
	{
	Sample *snd;

	snd=findSoundByName(name);
	if(!snd)
		{
		snd=load(name);

		if(!snd)
			{
			Log::fail("Sound.play: Nie udalo sie wczytac dzwieku \"%s\"", name.c_str());
			return -1;
			}
		}

	const int r=Mix_PlayChannel(channel, snd->snd, (repeat)?-1:0);
	Mix_Volume(r, volume*128);
	Log::debug("Sound.play: Odtwarzanie \"%s\" na kanale %d; vol %f; repeat: %s; pitch: %f; pan: %f", name.c_str(), r, volume, (repeat)?"TRUE":"FALSE", pitch, pan);

	return r;
	}
