/*
 * box.h
 *
 *  Created on: 19 kwi 2016
 *      Author: crm
 */


#pragma once

#include "asmmath.h"

#include <cmath>

#include "ray.h"

namespace Game
	{
	struct Box
		{
		Box(): pos(), size() {}
		Box(const AVector& pos, const AVector& size): pos(pos), size(size) {}

		Box operator+(const AVector& v) const {return Box(pos+v, size);}
		Box operator-(const AVector& v) const {return Box(pos-v, size);}

		operator bool() const {return size.x>0.0f && size.y>0.0f;}

		bool collision(const Box& b) const
			{
			const Box& a=*this;

			if(a.pos.x+a.size.x<b.pos.x ||
			   a.pos.x>b.pos.x+b.size.x ||
			   a.pos.y+a.size.y<b.pos.y ||
			   a.pos.y>b.pos.y+b.size.y)
				return false;
			return true;

			/*return std::abs(a.pos.x-b.pos.x)*2.0f < (a.size.x+b.size.x) &&
				   std::abs(a.pos.y-b.pos.y)*2.0f < (a.size.y+b.size.y);*/
			}

		bool collision(const Ray& r) const
			{
			const float drcpx=1.0f/r.dir.x;
			const float drcpy=1.0f/r.dir.y;

			float t1=(pos.x-r.pos.x)*drcpx;
			float t2=(pos.x+size.x-r.pos.x)*drcpx;
			float t3=(pos.y-r.pos.y)*drcpy;
			float t4=(pos.y+size.y-r.pos.y)*drcpy;

			float tmin=std::max( std::min(t1, t2), std::min(t3, t4));
			float tmax=std::min( std::max(t1, t2), std::max(t3, t4));

			return tmin<=tmax && tmax>0.0f;
			}

		AVector intersection(const Box& b) const
			{
			const Box& a=*this;

			return AVector(-std::min((a.pos.x+a.size.x-b.pos.x), (b.pos.x+b.size.x-a.pos.x)),
						   -std::min((a.pos.y+a.size.y-b.pos.y), (b.pos.y+b.size.y-a.pos.y)));
			}

		float distance(const Ray& r) const
			{
			const float drcpx=1.0f/r.dir.x;
			const float drcpy=1.0f/r.dir.y;

			float t1=(pos.x-r.pos.x)*drcpx;
			float t2=(pos.x+size.x-r.pos.x)*drcpx;
			float t3=(pos.y-r.pos.y)*drcpy;
			float t4=(pos.y+size.y-r.pos.y)*drcpy;

			float tmin=std::max( std::min(t1, t2), std::min(t3, t4));
			float tmax=std::min( std::max(t1, t2), std::max(t3, t4));

			if(tmin>tmax)
				return -1.0f;
			if(tmax<0.0f)
				return -1.0f;
			return tmin;
			}

		AVector pos;
		AVector size;
		};
	}
