/*
 * fade.cpp
 *
 *  Created on: 16 maj 2016
 *      Author: crm
 */

#include "fade.h"

#include "log.h"

#include "../engine/render/render.h"

using namespace Game;

bool Fade::init()
	{
	layers.reserve(4);

	return true;
	}

void Fade::update(float dt)
	{
	for(auto lay: layers)
		{
		if(!lay)
			continue;
		if(!lay->sprite)
			continue;

		lay->sprite.update(dt);

		if(lay->alpha==lay->target)
			continue;

		if(lay->alpha<lay->target)
			{
			lay->alpha+=dt;

			if(lay->alpha>lay->target)
				lay->alpha=lay->target;
			}
		else
			{
			lay->alpha-=dt;

			if(lay->alpha<lay->target)
				lay->alpha=lay->target;
			}
		}
	}

void Fade::print()
	{
	using namespace Engine;

	Engine::Render::pushState();
	Engine::Render::setMode(Engine::Render::MODE_GUI);

	//for(auto lay: layers)
	for(std::vector<Layer*>::reverse_iterator it=layers.rbegin(); it!=layers.rend(); ++it)
		{
		Layer *lay=*it;

		if(!lay)
			continue;
		if(lay->alpha<=0.0f)
			continue;
		if(!lay->sprite)
			continue;

		Render::setAlpha(lay->alpha);
		Render::print(lay->sprite, Orientation(AVector(0, 0), Orientation::GUI));
		}

	Engine::Render::popState();
	}

void Fade::clear()
	{
	for(auto lay: layers)
		delete lay;
	layers.clear();
	}


void Fade::set(unsigned layer, const Engine::SpritePtr& spr, float targetalpha, float alpha)
	{
	if(layer>10u)
		{
		Log::fail("Fade.set: Zbyt daleka warstwa (%u)", layer);
		return;
		}

	if(layer>=layers.size())
		{
		Log::info("Fade.set: Powiekszanie listy warstw (%u)", layer);

		const unsigned lsize=layers.size();

		layers.resize(layer+1);

		for(unsigned i=lsize; i<layers.size(); ++i)
			layers[i]=nullptr;
		}

	if(!layers[layer])
		{
		Log::info("Fade.set: Alokacja warstwy (%u)", layer);

		layers[layer]=new Layer();
		}

	if(alpha>=0.0f)
		layers[layer]->alpha=alpha;
	layers[layer]->target=targetalpha;
	layers[layer]->sprite=spr;
	}

void Fade::set(unsigned layer, float targetalpha, float alpha)
	{
	if(layer>10u)
		{
		Log::fail("Fade.set: Zbyt daleka warstwa (%u)", layer);
		return;
		}

	if(layer>=layers.size())
		{
		Log::info("Fade.set: Powiekszanie listy warstw (%u)", layer);

		layers.resize(layer+1);
		}

	if(!layers[layer])
		{
		Log::info("Fade.set: Alokacja warstwy (%u)", layer);

		layers[layer]=new Layer();
		}

	if(alpha>=0.0f)
		layers[layer]->alpha=alpha;
	layers[layer]->target=targetalpha;
	}

float Fade::getPct(unsigned layer) const
	{
	if(layer>=layers.size())
		return 0.0f;
	if(!layers[layer])
		return 0.0f;

	return layers[layer]->alpha;
	}
