/*
 * game.cpp
 *
 *  Created on: 8 maj 2016
 *      Author: crm
 */

#include "game.h"

#include "log.h"

#include "../consts.h"
#include "../engine/render/render.h"
#include "level.h"

using namespace Game;

const unsigned TEXT_HEIGHT=100;
const unsigned TEXT_MARGIN_V=2;
const unsigned TEXT_MARGIN_H=8;

const AVector CAM_OFFSET(0, 32);

bool Game::Game::init()
	{
	if(!sound.init(32, 44100))
		{
		Log::fail("Game.init: Nie udalo sie uruchomic dzwieku");
		return false;
		}

	if(!text.init(DEFAULT_FONT, "", Engine::Render::getScreenSize().x-2*TEXT_MARGIN_H, TEXT_HEIGHT-2.0f*TEXT_MARGIN_V))
		{
		Log::fail("Game.init: Nie udalo sie wczytac czcionki");
		return false;
		}

	text.update();

	cam=Engine::Camera(Engine::Camera::ORTHOGRAPHIC, Engine::Render::getScreenSize().x, Engine::Render::getScreenSize().y, -1000, 1000);

	dial.setGame(this);

	dialbg.setSprite("sprite/dial_bg.xml");

	if(!dialbg)
		{
		Log::fail("Game.init: Nie udalo sie wczytac grafiki tla dialogu");
		return false;
		}

	npcarrow.setSprite("sprite/npc_arrow.xml");

	if(!npcarrow)
		{
		Log::fail("Game.init: Nie udalo sie wczytac grafiki strzalki npc");
		return false;
		}

	return true;
	}

void Game::Game::update(float dt)
	{
	NPC *pnpc=lvl->getPlayer()->getDialogueNPC();

	lvl->update(dt);
	dial.update(dt);
	fade.update(dt);

	dialbg.update(dt);
	npcarrow.update(dt);

	if(dial.getMode()!=Dialogue::Mode::NONE)
		{
		if(text.getStr()!=dial.getMessage())
			{
			Log::debug("Game.update: \"%s\"", dial.getMessage().c_str());
			text.setStr(dial.getMessage());
			text.update();
			}

		if(dial.getMessage().length()>0)
			dialalpha=std::min(dialalpha+2.0f*dt, 1.0f);
		else
			dialalpha=std::max(dialalpha-1.0f*dt, 0.0f);

		if(!dial.isReady())
			dialbg.setAnimation(0);
		else if(dial.getMode()==Dialogue::Mode::NORMAL)
			dialbg.setAnimation(1);
		else
			dialbg.setAnimation(2);
		}
	else
		{
		dialalpha=std::max(dialalpha-1.0f*dt, 0.0f);
		camoverride=false;
		}

	if(!camoverride && lvl->getPlayer())
		{
		AVector target=lvl->getPlayer()->getPosition()-AVector(0, 0, 10)+CAM_OFFSET+CAM_OFFSET*fade.getPct(0);
		AVector diff=target-cam.getPos();
		float ddist=AVectorLength(diff).w;

		//cam.setZoom(1.0f+fade.getPct(0)*1.0f);

		//if(AVectorLength(diff).w>32.0f)
			{
			cam.move(AVectorNormalize(diff)*std::min(ddist*4.0f*dt, ddist));
			}

		/*const float MAX_DIST=32;

		if(cam.getPos().x>target.x+MAX_DIST)
			cam.setPos(AVector(target.x+MAX_DIST, cam.getPos().y));
		else if(cam.getPos().x<target.x-MAX_DIST)
			cam.setPos(AVector(target.x-MAX_DIST, cam.getPos().y));

		if(cam.getPos().y>target.y+MAX_DIST)
			cam.setPos(AVector(cam.getPos().x, target.y+MAX_DIST));
		else if(cam.getPos().y<target.y-MAX_DIST)
			cam.setPos(AVector(cam.getPos().x, target.y-MAX_DIST));*/
		}

	// Znikajaca strzalka
	if((dial.getMode()!=Dialogue::Mode::NONE || !lvl->getPlayer()->getDialogueNPC() || pnpc!=lvl->getPlayer()->getDialogueNPC()) && npcarrowvisible)
		{
		npcarrowvisible=false;

		Particle *part=new Particle();

		//part->init(npcarrow, 4.0f, AVector(0, 32.0f));
		part->init(npcarrow, Particle::Type::GRAVITY_MOVE, 2.0f);
		part->setSpeed(AVector(rand()/(RAND_MAX+1.0)*32-16, rand()/(RAND_MAX+1.0)*16)*4.0f);

		lvl->addParticle(part, npcarrowpos);
		}
	}

void Game::Game::print()
	{
	using namespace Engine;

	Render::setCamera(&cam);

	lvl->print();
	fade.print();

	if(dial.getMode()==Dialogue::Mode::NONE && lvl->getPlayer()->getDialogueNPC())
		{
		NPC *n=lvl->getPlayer()->getDialogueNPC();

		npcarrowpos=n->getPosition()+AVector(0, n->getBoundingBox().size.y, 10);
		npcarrowvisible=true;

		Render::print(npcarrow, Orientation(npcarrowpos));
		}

	if(dialalpha>0.0f)
		{
		Render::pushState();
		Render::setMode(Render::MODE_GUI);

		Render::setAlpha(dialalpha);
		Render::print(dialbg, Orientation(AVector(0, Render::getScreenSize().y-TEXT_HEIGHT), Orientation::GUI));
		text.print(Orientation(AVector(TEXT_MARGIN_H, Render::getScreenSize().y)));
		Render::setAlpha(1.0f);

		Render::popState();
		}
	}

void Game::Game::clear()
	{
	dial.clear();
	db.clear();

	dialalpha=0.0f;
	dialbg.clear();
	text.clear();

	npcarrow.clear();

	delete lvl;

	lvl=nullptr;
	}


bool Game::Game::setLevel(const std::string& path)
	{
	if(!lvl)
		{
		try
			{
			lvl=new Level();
			}
		catch(std::bad_alloc&)
			{
			Log::fail("Game.setLevel: Nie udalo sie utworzyc poziomu");
			return false;
			}
		}

	lvl->setGame(this);

	lvl->clear();

	if(!lvl->load(path))
		{
		Log::fail("Game.setLevel: Nie udalo sie wczytac poziomu \"%s\"", path.c_str());

		delete lvl;
		lvl=nullptr;

		return false;
		}

	if(lvl->getPlayer())
		cam.setPos(lvl->getPlayer()->getPosition()+CAM_OFFSET);
	else
		cam.setPos(AVector());

	Log::debug("Game.setLevel: Pozycja gracza: [%f, %f], pozycja kamery [%f, %f]", lvl->getPlayer()->getPosition().x, lvl->getPlayer()->getPosition().y, cam.getPos().x, cam.getPos().y);

	if(fade.getPct(0)>0) fade.set(0, 0.0f);
	if(fade.getPct(1)>0) fade.set(1, 0.0f);

	return true;
	}


bool Game::Game::enableShader(const std::string& path)
	{
	if(!shader.load(path))
		return false;

	senabled=true;

	Engine::Render::setEffect(shader);

	return true;
	}

void Game::Game::disableShader()
	{
	senabled=false;

	Engine::Render::unsetEffect();
	}
