/*
 * game.h
 *
 *  Created on: 8 maj 2016
 *      Author: crm
 */

#pragma once

#include <string>

#include "database.h"

#include "../engine/render/camera.h"
#include "../engine/render/shader.h"
#include "../engine/sound/sound.h"
#include "../engine/ui/text.h"
#include "dialogue.h"
#include "fade.h"

namespace Game
	{
	class Level;
	class Player;

	class Game
		{
		private:
			Level *lvl;

			Engine::Camera cam;
			Dialogue dial;
			Database db;
			Fade fade;
			Engine::Sound sound;

			float dialalpha;
			Engine::SpritePtr dialbg;
			Engine::UI::Text text;

			bool camoverride;

			bool senabled;
			Engine::Shader shader;

			AVector npcarrowpos;
			bool npcarrowvisible;
			Engine::SpritePtr npcarrow;

		public:
			Game(): lvl(nullptr), cam(), dial(), db(), fade(), sound(), dialalpha(0.0f), dialbg(), text(), camoverride(false), senabled(false), shader(), npcarrowpos(), npcarrowvisible(false), npcarrow()
				{
				}

			~Game()
				{
				clear();
				}

			bool init();
			void update(float dt);
			void print();
			void clear();

			bool setLevel(const std::string& path);

			bool enableShader(const std::string& path);
			void disableShader();

			Player* getPlayer() {return lvl?lvl->getPlayer():nullptr;}
			Level* getLevel() {return lvl;}
			Engine::Camera& getCamera() {return cam;}
			Dialogue& getDialogue() {return dial;}
			Database& getDatabase() {return db;}
			Fade& getFade() {return fade;}
			Engine::Shader& getShader() {return shader;}
			Engine::Sound& getSound() {return sound;}
		};
	}
