/*
 * level.h
 *
 *  Created on: 19 kwi 2016
 *      Author: crm
 */

#pragma once

#include <map>
#include <list>
#include <vector>
#include <string>

#include "asmmath.h"
#include "database.h"
#include "lua2.h"

#include "../engine/serializable.h"
#include "collider.h"
#include "object.h"
#include "backgroundobject.h"
#include "npc.h"
#include "player.h"
#include "particle.h"

namespace Game
	{
	class ObjectTemplate;
	class Game;

	class Level: public Engine::Serializable
		{
		private:
			AVector gravity;

			std::vector<Object*> objects; // Wszystkie obiekty
			std::vector<BackgroundObject*> bgobjs; // Obiekty tla (do kolizji)
			std::vector<NPC*> npcs; // NPC
			Player *player;

			std::list<Particle *> particles;

			std::vector<Collider> colliders; // Przeszkody

			Engine::SpritePtr skybox;

			std::string id;
			std::string path;
			std::map<std::string, ObjectTemplate*> templates; // Szablony obiektow

			std::string lpath; // Sciezka do skryptu LUA
			Lua lscript;

			Game *game;
			AVector startpos; // Pozycja startowa gracza

		public:
			Level(): player(nullptr), id(""), path(""), lpath(""), game(nullptr) {}
			virtual ~Level() {clear();}

			bool load(const std::string& path);
			bool save(const std::string& path);

			void update(float dt);
			void print();
			void clear();

			virtual char* serialize(int &size);
			virtual void deserialize(char *buff);

			// Funkcje do obslugi kolizji
			Collider test(const Box& b) const;

			// Edycja poziomu
			template<typename T> T* create(const std::string& tname, bool withColliders=true)
				{
				ObjectTemplate *tobj;

				if(templates.find(tname)==templates.end())
					return nullptr;

				tobj=templates[tname];

				T *obj;

				try
					{
					obj=new T(tobj, withColliders);
					}
				catch(std::bad_alloc& e)
					{
					return nullptr;
					}

				return obj;
				}

			bool addTemplate(const std::string& path);
			bool addTemplate(ObjectTemplate *tobj);
			bool addObject(Object *obj, const AVector& pos);
			bool addBackgroundObject(BackgroundObject *obj, const AVector& pos);
			bool addNPC(NPC *obj, const AVector& pos);
			bool addPlayer(Player *obj, const AVector& pos);
			bool addParticle(Particle *obj, const AVector& pos);

			void addCollider(const Collider& c);

			// Reszta
			void setGame(Game *s);
			void setGravity(const AVector& s) {gravity=s;}
			void setSkybox(const std::string& s) {skybox.setSprite(s);}

			AVector getGravity() const {return gravity;}
			std::string getID() const {return id;}
			std::string getPath() const {return path;}
			unsigned getNPCCount() const {return npcs.size();}
			NPC* getNPCByIndex(unsigned idx) const {return npcs[idx];}
			NPC* getNPCByID(const std::string& id) const;
			NPC* getNPCClosestTo(const AVector& pos, float maxDistance=-1.0f) const;
			Player* getPlayer() const {return player;}
			Lua& getLua() {return lscript;}
		};
	}
