/*
 * luafuncs.cpp
 *
 *  Created on: 16 maj 2016
 *      Author: crm
 */

#include "luafuncs.h"

#ifdef WIN32
	#define WINDOWS
#else
	#ifdef WIN64
		#define WINDOWS
	#else
		#define LINUX
	#endif
#endif

#ifdef WINDOWS
	#include <lua/lua.hpp>
	#include <lua/lualib.h>
	#include <lua/lauxlib.h>
#else
	#include <lua.hpp>
	#include <lualib.h>
	#include <lauxlib.h>
#endif

#include "log.h"
#include "resource.h"
#include "lua2.h"
#include "lua_misc.h"

#include "../consts.h"
#include "../engine/render/render.h"
#include "game.h"
#include "level.h"
#include "npc.h"
#include "particle.h"

#define ASSERT_ARGS(args) \
	if(lua_gettop(L)<(args))\
		{\
		Log::fail("LuaFuncs.%s: Za malo argumentow (otrzymano %d, oczekiwano %d)", __FUNCTION__, lua_gettop(L), (args));\
		return 0;\
		}
#define ASSERT_NPC(c) \
	if(c==NULL)\
		{\
		Log::fail("LuaFuncs.%s: NULL NPC", __FUNCTION__);\
		return 0;\
		}
#define ASSERT_LEVEL(c) \
	if(c==NULL)\
		{\
		Log::fail("LuaFuncs.%s: NULL level", __FUNCTION__);\
		return 0;\
		}
#define ASSERT_GAME(c) \
	if(c==NULL)\
		{\
		Log::fail("LuaFuncs.%s: NULL game", __FUNCTION__);\
		return 0;\
		}

//using namespace Game;

/**** Zwracanie zapisanych obiektow ****/
namespace LuaFuncs
	{
	Game::Game* getGame(lua_State *L);
	Game::Level* getLevel(lua_State *L);
	Game::NPC* getNPC(lua_State *L);
	}

// http://lua-users.org/lists/lua-l/2004-11/msg00440.html // nested namespaces

/**** Deklaracje ****/
namespace LuaFuncs
	{
	struct Function
		{
		Function(unsigned p, int(*ptr)(lua_State*), const char* name):
			permissions(p), ptr(ptr), name(name) {}

		unsigned permissions;
		int(*ptr)(lua_State*);
		const char* name;
		};
	}

namespace LuaFuncs
	{
// Bajery
	int test(lua_State *L);
	int getGlobalVal(lua_State *L);
	int getGlobalStr(lua_State *L);
	int setGlobalVal(lua_State *L);
	int setGlobalStr(lua_State *L);
	int include(lua_State *L);
	int dummy(lua_State *L);

// Poziom
	namespace LLevel
		{
		int getPlayerPosition(lua_State *L);
		int getPlayerSpeed(lua_State *L);
		int setPlayerPosition(lua_State *L);
		int getPlayerDirection(lua_State *L);
		int setPlayerDirection(lua_State *L);

		int setPlayerBounceLeft(lua_State *L);
		int setPlayerBounceRight(lua_State *L);

		int getCameraPosition(lua_State *L);
		int getScreenSize(lua_State *L);

		int getNPCCount(lua_State *L);
		int getNPCPositionByID(lua_State *L);
		int getNPCPosition(lua_State *L); // by index
		int setNpcPositionByID(lua_State *L);
		int setNpcPosition(lua_State *L); // by index
		int setNpcBounceLeftByID(lua_State *L);

		int setDialogue(lua_State *L);
		int setLevel(lua_State *L);

		int setShader(lua_State *L);
		int setShaderParam(lua_State *L);
		int unsetShader(lua_State *L);

		int addParticleStatic(lua_State *L);
		int addParticleLinear(lua_State *L);
		int addParticleGravity(lua_State *L);

		int getDBValue(lua_State *L);
		int setDBValue(lua_State *L);
		}

// NPC
	namespace LNPC
		{
		int getPosition(lua_State *L);
		int setPosition(lua_State *L);

		int moveLeft(lua_State *L);
		int moveRight(lua_State *L);
		int jump(lua_State *L);
		int bounceLeft(lua_State *L);
		int bounceRight(lua_State *L);

		int testCollision(lua_State *L);
		}
	}

namespace LuaFuncs
	{
	Function functions[]={
		//Function(NONE, test, "test"),
		Function(NPC|LEVEL, getGlobalVal, "getGlobalVal"),
		Function(NPC|LEVEL, getGlobalStr, "getGlobalStr"),
		Function(NPC|LEVEL, setGlobalVal, "setGlobalVal"),
		Function(NPC|LEVEL, setGlobalStr, "setGlobalStr"),
		Function(NPC|LEVEL, include, "include"),

		Function(LEVEL, LLevel::getPlayerPosition, "getPlayerPosition"),
		Function(LEVEL, LLevel::getPlayerSpeed, "getPlayerSpeed"),
		Function(LEVEL, LLevel::setPlayerPosition, "setPlayerPosition"),
		Function(LEVEL, LLevel::getPlayerDirection, "getPlayerDirection"),
		Function(LEVEL, LLevel::setPlayerDirection, "setPlayerDirection"),
		Function(LEVEL, LLevel::setPlayerBounceLeft, "setPlayerBounceLeft"),
		Function(LEVEL, LLevel::setPlayerBounceRight, "setPlayerBounceRight"),
		Function(LEVEL, LLevel::getCameraPosition, "getCameraPosition"),
		Function(LEVEL, LLevel::getScreenSize, "getScreenSize"),
		Function(LEVEL, LLevel::getNPCCount, "getNPCCount"),
		Function(LEVEL, LLevel::getNPCPositionByID, "getNPCPositionByID"),
		Function(LEVEL, LLevel::getNPCPosition, "getNPCPosition"),
		Function(LEVEL, LLevel::setNpcPositionByID, "setNpcPositionByID"),
		Function(LEVEL, LLevel::setNpcPosition, "setNpcPosition"),
		Function(LEVEL, LLevel::setNpcBounceLeftByID, "setNpcBounceLeftByID"),
		Function(LEVEL, LLevel::setDialogue, "setDialogue"),
		Function(LEVEL, LLevel::setLevel, "setLevel"),
		Function(LEVEL, LLevel::setShader, "setShader"),
		Function(LEVEL, LLevel::setShaderParam, "setShaderParam"),
		Function(LEVEL, LLevel::unsetShader, "unsetShader"),
		Function(LEVEL, LLevel::addParticleStatic, "addParticleStatic"),
		Function(LEVEL, LLevel::addParticleLinear, "addParticleLinear"),
		Function(LEVEL, LLevel::addParticleGravity, "addParticleGravity"),
		Function(LEVEL, LLevel::getDBValue, "getDBValue"),
		Function(LEVEL, LLevel::setDBValue, "setDBValue"),

		Function(NPC, LNPC::getPosition, "getPosition"),
		Function(NPC, LNPC::setPosition, "setPosition"),
		Function(NPC, LNPC::moveLeft, "moveLeft"),
		Function(NPC, LNPC::moveRight, "moveRight"),
		Function(NPC, LNPC::jump, "jump"),
		Function(NPC, LNPC::bounceLeft, "bounceLeft"),
		Function(NPC, LNPC::bounceRight, "bounceRight"),
		Function(NPC, LNPC::testCollision, "testCollision"),
		};
	}

/**** Rejestrowanie ****/
namespace LuaFuncs
	{
	void load(Lua& lua, unsigned permissions)
		{
		lua.addGlobal("test", test);

		const unsigned size=sizeof(functions)/sizeof(functions[0]);
		for(unsigned i=0; i<size; ++i)
			{
			const Function &f=functions[i];

			if((permissions&f.permissions)==f.permissions)	// ma dostep
				lua.addGlobal(f.name, f.ptr);
			else
				lua.addGlobal(f.name, dummy);
			}
		}

	Game::Game* getGame(lua_State *L)
		{
		lua_pushstring(L, LUA_PTR_GAME);
		lua_gettable(L, LUA_REGISTRYINDEX);

		Game::Game *a=static_cast<Game::Game *>(lua_touserdata(L, -1));

		lua_settop(L, -2);

		return a;
		}

	Game::Level* getLevel(lua_State *L)
		{
		lua_pushstring(L, LUA_PTR_LEVEL);
		lua_gettable(L, LUA_REGISTRYINDEX);

		Game::Level *a=static_cast<Game::Level *>(lua_touserdata(L, -1));

		lua_settop(L, -2);

		return a;
		}

	Game::NPC* getNPC(lua_State *L)
		{
		lua_pushstring(L, LUA_PTR_NPC);
		lua_gettable(L, LUA_REGISTRYINDEX);

		Game::NPC *a=static_cast<Game::NPC *>(lua_touserdata(L, -1));

		lua_settop(L, -2);

		return a;
		}
	}

/**** Funkcje do dzialan specjalnych ****/
namespace LuaFuncs
	{
	int test(lua_State *L)
		{
		if(!L)
			return 0;
		printf("LuaFuncs.lTest");
		return 0;
		}

	int getGlobalVal(lua_State *L)
		{
		if(!L)
			return 0;
		if(lua_gettop(L)<1)
			{
			printf("FAIL: LuaFuncs.%s: Za malo argumentow", __FUNCTION__);
			return 0;
			}
		lua_pushstring(L, lua_tostring(L, 1));
		lua_gettable(L, LUA_REGISTRYINDEX);
		lua_pushnumber(L, lua_tonumber(L, -1));
		return 1;
		}

	int getGlobalStr(lua_State *L)
		{
		if(!L)
			return 0;
		if(lua_gettop(L)<1)
			{
			printf("FAIL: LuaFuncs.%s: Za malo argumentow", __FUNCTION__);
			return 0;
			}
		lua_pushstring(L, lua_tostring(L, 1));
		lua_gettable(L, LUA_REGISTRYINDEX);
		lua_pushstring(L, lua_tostring(L, -1));
		return 1;
		}

	int setGlobalVal(lua_State *L)
		{
		if(!L)
			return 0;
		if(lua_gettop(L)<2)
			{
			printf("FAIL: LuaFuncs.%s: Za malo argumentow", __FUNCTION__);
			return 0;
			}
		const char* name=lua_tostring(L, 1);
		int val=lua_tonumber(L, 2);
		lua_pushstring(L, name);
		lua_pushnumber(L, val);
		lua_settable(L, LUA_REGISTRYINDEX);
		return 0;
		}

	int setGlobalStr(lua_State *L)
		{
		if(!L)
			return 0;
		if(lua_gettop(L)<2)
			{
			printf("FAIL: LuaFuncs.%s: Za malo argumentow", __FUNCTION__);
			return 0;
			}
		const char* name=lua_tostring(L, 1);
		const char* val=lua_tostring(L, 2);
		lua_pushstring(L, name);
		lua_pushstring(L, val);
		lua_settable(L, LUA_REGISTRYINDEX);
		return 0;
		}

	int include(lua_State *L)
		{
		if(!L)
			return 0;
		if(lua_gettop(L)<1)
			{
			printf("FAIL: LuaFuncs.%s: Za malo argumentow", __FUNCTION__);
			return 0;
			}

		char *script;

		script=Resource::load(lua_tostring(L, 1));

		if(script==NULL)
			{
			printf("FAIL: LuaFuncs.%s: Nie znaleziono pliku \"%s\"", __FUNCTION__, lua_tostring(L, 1));
			return 0;
			}
		if(luaL_loadbufferx(L, script, Resource::getSize(), "include", NULL)!=LUA_OK)
			{
			printf("FAIL: LuaFuncs.%s: Problem przy ladowaniu skryptu; %s", __FUNCTION__, lua_tostring(L, -1));
			return 0;
			}
		if(lua_pcallk(L, 0, LUA_MULTRET, 0, 0, NULL)!=LUA_OK)
			{
			printf("FAIL: LuaFuncs.%s: Problem przy wykonaniu skryptu; %s", __FUNCTION__, lua_tostring(L, -1));
			return 0;
			}

		delete [] script;

		return 0;
		}

	int dummy(lua_State *)
		{
		return 0;
		}
	}

namespace LuaFuncs
	{
	/**** Level ****/
	namespace LLevel
		{
		int getPlayerPosition(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(0);

			if(!lvl->getPlayer())
				return 0;

			AVector p=lvl->getPlayer()->getPosition();

			lua_pushnumber(L, p.x);
			lua_pushnumber(L, p.y);

			return 2;
			}

		int getPlayerSpeed(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(0);

			if(!lvl->getPlayer())
				return 0;

			AVector p=lvl->getPlayer()->getSpeed();

			lua_pushnumber(L, p.x);
			lua_pushnumber(L, p.y);

			return 2;
			}


		int setPlayerPosition(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(2);

			if(!lvl->getPlayer())
				return 0;

			lvl->getPlayer()->setPosition(AVector(lua_tonumber(L, -2), lua_tonumber(L, -1)));

			return 0;
			}

		int getPlayerDirection(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(0);

			if(!lvl->getPlayer())
				return 0;

			if(lvl->getPlayer()->getSprite().getDirectionID()==(unsigned)Engine::Sprite::DirectionName::RIGHT)
				lua_pushinteger(L, 1);
			else
				lua_pushinteger(L, -1);

			return 1;
			}

		int setPlayerDirection(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(1);

			if(!lvl->getPlayer())
				return 0;

			if(lua_tointeger(L, -1)>0)
				lvl->getPlayer()->getSprite().setDirection((unsigned)Engine::Sprite::DirectionName::RIGHT);
			else
				lvl->getPlayer()->getSprite().setDirection((unsigned)Engine::Sprite::DirectionName::LEFT);

			return 0;
			}


		int setPlayerBounceLeft(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(2);

			if(!lvl->getPlayer())
				return 0;

			lvl->getPlayer()->bounceLeft(lua_tonumber(L, -2), lua_tonumber(L, -1));

			return 0;
			}

		int setPlayerBounceRight(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(0);

			if(!lvl->getPlayer())
				return 0;

			lvl->getPlayer()->bounceRight(lua_tonumber(L, -2), lua_tonumber(L, -1));

			return 0;
			}


		int getCameraPosition(lua_State *L)
			{
			Game::Game *game=getGame(L);

			ASSERT_GAME(game);
			ASSERT_ARGS(0);

			lua_pushnumber(L, game->getCamera().getPos().x);
			lua_pushnumber(L, game->getCamera().getPos().y);

			return 2;
			}

		int getScreenSize(lua_State *L)
			{
			lua_pushnumber(L, Engine::Render::getScreenSize().x);
			lua_pushnumber(L, Engine::Render::getScreenSize().y);

			return 2;
			}


		int getNPCCount(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(0);

			lua_pushinteger(L, lvl->getNPCCount());

			return 1;
			}

		int getNPCPositionByID(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(1);

			Game::NPC *npc=lvl->getNPCByID(lua_tostring(L, -1));

			if(!npc)
				return 0;

			lua_pushnumber(L, npc->getPosition().x);
			lua_pushnumber(L, npc->getPosition().y);

			return 2;
			}

		int getNPCPosition(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(1);

			unsigned idx=lua_tointeger(L, -1);

			if(idx>lvl->getNPCCount())
				return 0;

			Game::NPC *npc=lvl->getNPCByIndex(idx);

			if(!npc)
				return 0;

			lua_pushnumber(L, npc->getPosition().x);
			lua_pushnumber(L, npc->getPosition().y);

			return 2;
			}

		int setNpcPositionByID(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(3);

			Game::NPC *npc=lvl->getNPCByID(lua_tostring(L, -3));

			if(!npc)
				return 0;

			npc->setPosition(AVector(lua_tonumber(L, -2), lua_tonumber(L, -1)));

			return 0;
			}

		int setNpcPosition(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(3);

			unsigned idx=lua_tointeger(L, -3);

			if(idx>lvl->getNPCCount())
				return 0;

			Game::NPC *npc=lvl->getNPCByIndex(idx);

			if(!npc)
				return 0;

			npc->setPosition(AVector(lua_tonumber(L, -2), lua_tonumber(L, -1)));

			return 0;
			}

		int setNpcBounceLeftByID(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(3);

			Game::NPC *npc=lvl->getNPCByID(lua_tostring(L, -3));

			if(!npc)
				return 0;

			npc->bounceLeft(lua_tonumber(L, -2), lua_tonumber(L, -1));

			return 0;
			}


		int setDialogue(lua_State *L)
			{
			Game::Game *game=getGame(L);

			ASSERT_GAME(game);
			ASSERT_ARGS(1);

			if(!game->getDialogue().init(lua_tostring(L, -1)))
				{
				Log::fail("LuaFuncs.%s: Nie udalo sie uruchomic dialogu", __FUNCTION__);
				return 0;
				}

			if(game->getPlayer())
				game->getPlayer()->setPlayerModeTalking();

			return 0;
			}

		int setLevel(lua_State *L)
			{
			Game::Game *game=getGame(L);

			ASSERT_GAME(game);
			ASSERT_ARGS(1);

			if(!game->setLevel(lua_tostring(L, -1)))
				{
				Log::fail("LuaFuncs.%s: Nie udalo sie wczytac poziomu", __FUNCTION__);
				return 0;
				}

			return 0;
			}


		int setShader(lua_State *L)
			{
			Game::Game *game=getGame(L);

			ASSERT_GAME(game);
			ASSERT_ARGS(1);

			if(!game->enableShader(lua_tostring(L, -1)))
				{
				Log::fail("LuaFuncs.%s: Nie udalo sie uruchomic shadera", __FUNCTION__);
				return 0;
				}

			return 0;
			}

		int setShaderParam(lua_State *L)
			{
			Game::Game *game=getGame(L);

			ASSERT_GAME(game);
			ASSERT_ARGS(2);

			game->getShader().uniform(lua_tostring(L, -2), (float)lua_tonumber(L, -1));

			return 0;
			}

		int unsetShader(lua_State *L)
			{
			Game::Game *game=getGame(L);

			ASSERT_GAME(game);
			ASSERT_ARGS(0);

			game->disableShader();

			return 0;
			}

		int addParticleStatic(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(4);

			// Zakladam, ze partikle zawsze uda sie zrobic
			Game::Particle *part=new Game::Particle();

			part->init((Engine::SpritePtr)lua_tostring(L, -1), Game::Particle::Type::STATIC, lua_tonumber(L, -2));

			lvl->addParticle(part, AVector(lua_tonumber(L, -4), lua_tonumber(L, -3)));

			return 0;
			}

		int addParticleLinear(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(6);

			Game::Particle *part=new Game::Particle();

			part->init((Engine::SpritePtr)lua_tostring(L, -1), lua_tonumber(L, -2), AVector(lua_tonumber(L, -4), lua_tonumber(L, -3)));

			lvl->addParticle(part, AVector(lua_tonumber(L, -6), lua_tonumber(L, -5)));

			return 0;
			}

		int addParticleGravity(lua_State *L)
			{
			Game::Level *lvl=getLevel(L);

			ASSERT_LEVEL(lvl);
			ASSERT_ARGS(6);

			Game::Particle *part=new Game::Particle();

			part->init((Engine::SpritePtr)lua_tostring(L, -1), Game::Particle::Type::GRAVITY_MOVE, lua_tonumber(L, -2));
			part->setSpeed(AVector(lua_tonumber(L, -4), lua_tonumber(L, -3)));

			lvl->addParticle(part, AVector(lua_tonumber(L, -6), lua_tonumber(L, -5)));

			return 0;
			}


		int getDBValue(lua_State *L)
			{
			Game::Game *game=getGame(L);

			ASSERT_GAME(game);
			ASSERT_ARGS(1);

			lua_pushinteger(L, game->getDatabase().getVal(lua_tostring(L, -1)));

			return 1;
			}

		int setDBValue(lua_State *L)
			{
			Game::Game *game=getGame(L);

			ASSERT_GAME(game);
			ASSERT_ARGS(2);

			game->getDatabase().setVal(lua_tostring(L, -2), lua_tointeger(L, -1));

			return 0;
			}
		}

	/**** Level ****/
	namespace LNPC
		{
		int getPosition(lua_State *L)
			{
			Game::NPC *npc=getNPC(L);

			ASSERT_NPC(npc);
			ASSERT_ARGS(0);

			lua_pushnumber(L, npc->getPosition().x);
			lua_pushnumber(L, npc->getPosition().y);

			return 2;
			}

		int setPosition(lua_State *L)
			{
			Game::NPC *npc=getNPC(L);

			ASSERT_NPC(npc);
			ASSERT_ARGS(2);

			npc->setPosition(AVector(lua_tonumber(L, -2), lua_tonumber(L, -1)));

			return 0;
			}


		int moveLeft(lua_State *L)
			{
			Game::NPC *npc=getNPC(L);

			ASSERT_NPC(npc);
			ASSERT_ARGS(0);

			npc->moveLeft();

			return 0;
			}

		int moveRight(lua_State *L)
			{
			Game::NPC *npc=getNPC(L);

			ASSERT_NPC(npc);
			ASSERT_ARGS(0);

			npc->moveRight();

			return 0;
			}

		int jump(lua_State *L)
			{
			Game::NPC *npc=getNPC(L);

			ASSERT_NPC(npc);
			ASSERT_ARGS(1);

			npc->jump(lua_tonumber(L, -1));

			return 0;
			}

		int bounceLeft(lua_State *L)
			{
			Game::NPC *npc=getNPC(L);

			ASSERT_NPC(npc);
			ASSERT_ARGS(2);

			npc->bounceLeft(lua_tonumber(L, -2), lua_tonumber(L, -1));

			return 0;
			}

		int bounceRight(lua_State *L)
			{
			Game::NPC *npc=getNPC(L);

			ASSERT_NPC(npc);
			ASSERT_ARGS(1);

			npc->bounceRight(lua_tonumber(L, -2), lua_tonumber(L, -1));

			return 0;
			}


		int testCollision(lua_State *L)
			{
			Game::NPC *npc=getNPC(L);

			ASSERT_NPC(npc);
			ASSERT_ARGS(4);

			if(!npc->getLevel())
				return 0;

			bool r=npc->getLevel()->test(Game::Box(AVector(lua_tonumber(L, -4), lua_tonumber(L, -3)), AVector(lua_tonumber(L, -2), lua_tonumber(L, -1))));

			lua_pushboolean(L, r);

			return 1;
			}
		}
	}
