/*
 * npc.cpp
 *
 *  Created on: 1 maj 2016
 *      Author: crm
 */

#include "npc.h"

#include "log.h"

#include "../consts.h"
#include "luafuncs.h"

using namespace Game;

NPC::NPC(): ActiveObject(), id(""), flags(FLAG_NONE), mode(NPCMode::NORMAL), lpath(""), lscript(), spath("")
	{
	//
	}

NPC::NPC(const ObjectTemplate *tobj, bool withColliders): ActiveObject(tobj, withColliders), id(""), flags(FLAG_NONE), mode(NPCMode::NORMAL), lpath(""), lscript(), spath("")
	{
	lscript.init();
	}


bool NPC::init(const std::string& id, const std::string& spath, const std::string& lpath)
	{
	this->id=id;
	this->spath=spath;
	this->lpath=lpath;

	flags=FLAG_NONE;

	Log::info("NPC.init: [NPC:%s] Startowanie NPC, skrypt \"%s\", lua \"%s\"", id.c_str(), spath.c_str(), lpath.c_str());

	if(lpath.empty())
		{
		Log::info("NPC.init: [NPC:%s] Nie zdefiniowano skryptu Lua", id.c_str());
		}
	else if(!lscript.open(lpath))
		{
		Log::fail("NPC.init: [NPC:%s] Nie udalo sie zaladowac skryptu Lua \"%s\"", id.c_str(), lpath.c_str());
		return false;
		}
	else
		{
		Log::ok("NPC.init: [NPC:%s] Zaladowano skrypt Lua \"%s\"", id.c_str(), lpath.c_str());
		flags|=FLAG_HAS_AI;

		lscript.addRegistry(LUA_PTR_LEVEL, level);
		lscript.addRegistry(LUA_PTR_NPC, this);
		lscript.addRegistry(LUA_PTR_GAME, game);

		LuaFuncs::load(lscript, LuaFuncs::NPC|LuaFuncs::LEVEL);

		lscript.call(LUA_INIT);
		}

	if(!spath.empty())
		{
		Log::ok("NPC.init: [NPC:%s] NPC ma skrypt \"%s\"", id.c_str(), spath.c_str());

		flags|=FLAG_HAS_DIALOG;
		}

	return true;
	}

void NPC::update(float dt)
	{
	if(flags&FLAG_DISABLED)
		return;

	if(flags&FLAG_HAS_AI)
		{
		lscript.addGlobal("dt", dt);
		lscript.call(LUA_UPDATE);
		}

	if(isUnmoving())
		{
		if(mode==NPCMode::NORMAL)
			updateAnim();
		else
			setAnimation("talking");

		Object::update(dt);

		movstate=MOVE_NONE;
		}
	else
		{
		if(mode!=NPCMode::NORMAL)
			setAnimation("talking");

		ActiveObject::update(dt);
		}
	}

void NPC::print()
	{
	if(flags&FLAG_DISABLED)
		return;

	ActiveObject::print();
	}

void NPC::clear()
	{
	ActiveObject::clear();
	}


Collider NPC::test(const Box& b) const
	{
	if(isDisabled())
		return Collider();

	return Object::test(b);
	}


char* NPC::serialize(int &size)
	{
	return nullptr;
	}

void NPC::deserialize(char *buff)
	{
	//
	}
