/*
 * npc.h
 *
 *  Created on: 1 maj 2016
 *      Author: crm
 */

#pragma once

#include <string>

#include "lua2.h"

#include "activeobject.h"

namespace Game
	{
	class NPC: public ActiveObject
		{
		protected:
			enum NPCFlag
				{
				FLAG_NONE=       0x0,
				FLAG_DISABLED=   0x1<<0,
				FLAG_HAS_DIALOG= 0x1<<1,
				FLAG_HAS_AI=     0x1<<2,
				FLAG_UNMOVING=   0x1<<3,
				};

			enum class NPCMode
				{
				NORMAL,
				TALKING,
				};

			std::string id;

			unsigned flags;
			NPCMode mode;

			std::string lpath; // Sciezka do skryptu LUA
			Lua lscript;

			std::string spath; // Sciezka do skryptu dialogowego

		public:
			NPC();
			NPC(const ObjectTemplate *tobj, bool withColliders=true);

			bool init(const std::string& id, const std::string& spath, const std::string& lpath);

			virtual void update(float dt);
			virtual void print();
			virtual void clear();

			virtual Collider test(const Box& b) const;

			virtual char* serialize(int &size);
			virtual void deserialize(char *buff);

			void startDialogue() {mode=NPCMode::TALKING;}
			void stopDialogue() {mode=NPCMode::NORMAL; setAnimation(0u);}

			std::string getID() const {return id;}
			std::string getScriptPath() const {return spath;}
			std::string getLuaPath() const {return lpath;}
			bool hasDialog() const {return flags&FLAG_HAS_DIALOG;}
			bool hasAI() const {return flags&FLAG_HAS_AI;}
			bool isUnmoving() const {return flags&FLAG_UNMOVING;}
			bool isDisabled() const {return flags&FLAG_DISABLED;}

			void setDisabled(bool s) {flags=s?(flags|FLAG_DISABLED):(flags&~FLAG_DISABLED);}
			void setUnmoving(bool s) {flags=s?(flags|FLAG_UNMOVING):(flags&~FLAG_UNMOVING);}
		};
	}
