/*
 * object.h
 *
 *  Created on: 26 kwi 2016
 *      Author: crm
 */

#pragma once

#include <vector>

#include "../engine/serializable.h"
#include "../engine/orientation.h"
#include "../engine/render/spriteptr.h"

#include "box.h"
#include "collider.h"

namespace Game
	{
	class Level;
	class ObjectTemplate;
	class Game;

	class Object: public Engine::Serializable
		{
		protected:
			AVector pos;

			Box boundingbox;
			std::vector<Collider> colliders;

			Level *level;
			Game *game;
			const ObjectTemplate *tobj;

			Engine::SpritePtr sprite;

		public:
			Object(): pos(), boundingbox(), level(nullptr), game(nullptr), tobj(nullptr), sprite() {}
			Object(const ObjectTemplate *tobj, bool withColliders=true);
			virtual ~Object() {clear();}

			virtual void update(float dt);
			virtual void print();
			virtual void clear();

			virtual char* serialize(int &size)=0;
			virtual void deserialize(char *buff)=0;

			virtual Collider test(const Box& b) const;

			AVector getPosition() const {return pos;}
			Box getBoundingBox() const {return boundingbox+pos;}
			Level* getLevel() const {return level;}
			const ObjectTemplate* getTemplate() const {return tobj;}
			Engine::SpritePtr& getSprite() {return sprite;}
			unsigned getAnimation() const {return sprite.getAnimationID();}
			unsigned getAnimationDirection() const {return sprite.getDirectionID();}
			bool hasColliders() const {return !colliders.empty();}

			void setPosition(const AVector& s) {pos=s;}
			void setLevel(Level& s) {level=&s;}
			void setGame(Game* s) {game=s;}
			void setSprite(const Engine::SpritePtr& s) {sprite=s;}
			void setAnimation(const std::string& s) {sprite.setAnimation(s);}
			void setAnimation(unsigned s) {sprite.setAnimation(s);}
			void setAnimationDirection(unsigned s) {sprite.setDirection(s);}
		};
	}
