/*
 * particle.cpp
 *
 *  Created on: 14 maj 2016
 *      Author: crm
 */

#include "particle.h"

#include "log.h"

#include "../engine/render/render.h"
#include "box.h"
#include "level.h"
#include "object.h"

using namespace Game;

bool Particle::init(const Engine::SpritePtr& sprite, Type type, float ttl)
	{
	this->sprite=sprite;
	this->type=type;
	this->ttl=ttl;

	switch(type)
		{
		case Type::FOLLOW_OBJECT:
			Log::fail("Particle.init: Zly init dla typu FOLLOW_OBJECT");
			return false;

		default:
		case Type::STATIC:
		case Type::GRAVITY_MOVE:
		case Type::LINEAR_MOVE:
			return true;
		}
	}

bool Particle::init(const Engine::SpritePtr& sprite, float ttl, AVector offset, Object *target)
	{
	this->sprite=sprite;
	this->type=Type::FOLLOW_OBJECT;
	this->ttl=ttl;

	this->offset=offset;

	return true;
	}

bool Particle::init(const Engine::SpritePtr& sprite, float ttl, AVector spd)
	{
	this->sprite=sprite;
	this->type=Type::LINEAR_MOVE;
	this->ttl=ttl;
	this->spd=spd;

	return true;
	}

void Particle::update(float dt)
	{
	if(ttl>0.0f)
		{
		ttl-=dt;

		/*if(ttl<0.0f)
			ttl=0.0f;*/
		}

	sprite.update(dt);

	switch(type)
		{
		case Type::FOLLOW_OBJECT:
			pos=target->getPosition()+offset;
		break;

		case Type::LINEAR_MOVE:
			pos=pos+spd*dt;
		break;

		case Type::GRAVITY_MOVE:
			spd=spd+lvl->getGravity()*dt;
			pos=pos+spd*dt;

			if(spd.x<0.0f && lvl->test(Box(pos+AVector(-2.0, -1.0), AVector(1, 2))))
				spd.x*=-1;
			else if(spd.x>0.0f && lvl->test(Box(pos+AVector( 2.0, -1.0), AVector(1, 2))))
				spd.x*=-1;
			if(spd.y<0.0f && lvl->test(Box(pos+AVector(-1.0, -2.0), AVector(2, 1))))
				spd.y*=-1;
			else if(spd.y<0.0f && lvl->test(Box(pos+AVector(-1.0,  2.0), AVector(2, 1))))
				spd.y*=-1;
		break;

		default:
		case Type::STATIC:
			// noop
		break;
		}
	}

void Particle::print()
	{
	if(ttl>=1.0f)
		{
		Engine::Render::print(sprite, Engine::Orientation(pos, Engine::Orientation::FLAT_XY));
		}
	else
		{
		Engine::Render::setAlpha(ttl);
		Engine::Render::print(sprite, Engine::Orientation(pos, Engine::Orientation::FLAT_XY));
		Engine::Render::setAlpha(1.0f);
		}
	}

void Particle::clear()
	{
	sprite.clear();
	}
