/*
 * player.cpp
 *
 *  Created on: 7 maj 2016
 *      Author: crm
 */

#include "player.h"

#include "log.h"

#include "game.h"
#include "level.h"
#include "dialogue.h"

using namespace Game;

Player::Player(): ActiveObject(), mode(PlayerMode::NORMAL), dialnpc(nullptr)
	{
	//
	}

Player::Player(const ObjectTemplate *tobj, bool withColliders): ActiveObject(tobj, withColliders), mode(PlayerMode::NORMAL), dialnpc(nullptr)
	{
	//
	}


bool Player::init()
	{
	return ctrl.init("controls", "player_01");
	}

void Player::update(float dt)
	{
	Dialogue &dial=game->getDialogue();

	ctrl.update(dt);

	switch(mode)
		{
		case PlayerMode::TALKING:
			ActiveObject::updateMove(dt);
			Object::update(dt);

			setAnimation("talking");

			if(dial.getMode()==Dialogue::Mode::NONE)
				{
				if(dialnpc)
					dialnpc->stopDialogue();

				mode=PlayerMode::NORMAL;
				break;
				}

			if(ctrl.onKeyDown(Controller::CONFIRM))
				dial.next();
			else if(ctrl.onKeyDown(Controller::NEXT))
				dial.choiceNext();
			else if(ctrl.onKeyDown(Controller::PREVIOUS))
				dial.choicePrevious();
		break;

		default:
		case PlayerMode::NORMAL:
			{
			if(dial.getMode()!=Dialogue::Mode::NONE)
				{
				mode=PlayerMode::TALKING;
				break;
				}

			dialnpc=level->getNPCClosestTo(pos, 64.0f);

			if(dialnpc /*&& AVectorLength(spd+mspd).w<=1.0f*/ && std::abs(spd.y)<=0.1f)
				{
				if(ctrl.onKeyDown(Controller::CONFIRM) && dial.init(dialnpc->getScriptPath()))
					{
					dialnpc->startDialogue();

					Log::info("Player.update: Zaczynanie dialogu z \"%s\", skrypt \"%s\"", dialnpc->getID().c_str(), dialnpc->getScriptPath().c_str());
					mode=PlayerMode::TALKING;
					}
				}
			else
				dialnpc=nullptr;

			if(ctrl.isPressed(Controller::LEFT))
				moveLeft();
			if(ctrl.isPressed(Controller::RIGHT))
				moveRight();
			if(ctrl.isPressed(Controller::JUMP))
				jump(64.0f);

			ActiveObject::update(dt);
			}
		break;
		}
	}

void Player::print()
	{
	ActiveObject::print();
	}

void Player::clear()
	{
	ActiveObject::clear();
	}


char* Player::serialize(int &size)
	{
	return nullptr;
	}

void Player::deserialize(char *buff)
	{
	//
	}

