/*
 * asmmath.cpp
 *
 *  Created on: 18-03-2012
 *      Author: crm
 */

#include "asmmath.h"

#include <cmath>

#define ASMMATH_NO_SSE41

void nop()
	{
	asm (
		"nop\n\t"
		);
	}

void sincos(float val, float& s, float& c)
	{
	asm volatile(
		"fsincos\n\t"
		: "=t" (c), "=u" (s)		// t-> gora stosu fpu, u-> drugi od gory element stosu fpu
		: "0" (val)					// fsincos zwaraca wynik do ST0	// A nie raczej 'pobiera argument' z ST0?
		);
	}

/******************************************************************************/
/****************** Operacje wektorowe, woooooo~ ******************************/
/******************************************************************************/
AVector AVectorAdd(const AVector& a, const AVector& b)
	{
	AVector c;

	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"movups (%2), %%xmm1\n\t"
		"addps %%xmm1, %%xmm0\n\t"
		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a), "r"(&b)
		);
	return c;

	/*AVector c;
	c=a.x+b.x;
	c=a.y+b.y;
	c=a.z+b.z;
	c=a.w+b.w;
	return c;*/
	}

AVector AVectorSubtract(const AVector& a, const AVector& b)
	{
	AVector c;

	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"movups (%2), %%xmm1\n\t"
		"subps %%xmm1, %%xmm0\n\t"
		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a), "r"(&b)
		);

	return c;

	/*AVector c;
	c=a.x-b.x;
	c=a.y-b.y;
	c=a.z-b.z;
	c=a.w-b.w;
	return c;*/
	}

AVector AVectorMultiply(const AVector& a, const AVector& b)
	{
	AVector c;

	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"movups (%2), %%xmm1\n\t"
		"mulps %%xmm1, %%xmm0\n\t"
		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a), "r"(&b)
		);

	return c;

	/*AVector c;
	c=a.x*b.x;
	c=a.y*b.y;
	c=a.z*b.z;
	c=a.w*b.w;
	return c;*/
	}

AVector AVectorDivide(const AVector& a, const AVector& b)
	{
	AVector c;

	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"movups (%2), %%xmm1\n\t"
		"divps %%xmm1, %%xmm0\n\t"
		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a), "r"(&b)
		);

	return c;

	/*AVector c;
	c=a.x/b.x;
	c=a.y/b.y;
	c=a.z/b.z;
	c=a.w/b.w;
	return c;*/
	}

/*AVector AVectorShuffle(AVector a, char flag)
	{*/
	/* Piekna rozpiska co i jak~
	 * const unsigned char BCAST_X=0x00;
	 * const unsigned char BCAST_Y=0x55;
	 * const unsigned char BCAST_Z=0xAA;
	 * const unsigned char BCAST_W=0xFF;
	 * const unsigned char REVERSE=0x1B;
	 * const unsigned char ROTLEFT=0x39;
	 * const unsigned char ROTRIGH=0x93;
	 */

	/*asm volatile(
		"movups (%1), %%xmm0\n\t"
		//"mov (%2), ecx\n\t"
		"shufps %2, %%xmm0, %%xmm0\n\t"
		"movups %%xmm0, %0\n\t"
		: "=m"(a)
		: "r"(&a), "m"(&flag)
		);

	return a;
	}*/

AVector AVectorNormalize(const AVector& a)
	{
	AVector c;

	/*asm volatile(
	// Ulepszona? wersja, 20120830m 10:25
		"movups (%1), %%xmm0\n\t"
		"movaps %%xmm0, %%xmm2\n\t"
		"mulps %%xmm0, %%xmm0\n\t"			// v^2
		"movaps %%xmm0, %%xmm1\n\t"
		"shufps $0xB1, %%xmm1, %%xmm1\n\t"	// 0xB1 = 10110001
		"addps %%xmm0, %%xmm1\n\t"			// xmm1 -> (x, y, z, w)+(y, x, w, z)=(x+y, x+y, z+w, z+w)
		"movaps %%xmm1, %%xmm0\n\t"
		"shufps $0x0A, %%xmm1, %%xmm1\n\t"	// 0x0A = 00001010
		"addps %%xmm0, %%xmm1\n\t"			// xmm1 -> (x+y+z+w, , , )
		"rsqrtps %%xmm1, %%xmm1\n\t"
		"mulps %%xmm2, %%xmm1\n\t"
		"movups %%xmm1, %0\n\t"
		: "=m"(c)
		: "r"(&a)
		);*/

	const double len=sqrt(a.x*a.x+a.y*a.y+a.z*a.z+a.w*a.w);

	c.x=a.x/len;
	c.y=a.y/len;
	c.z=a.z/len;
	c.w=a.w/len;

	return c;
	}

AVector AVectorLength(const AVector& a)
	{
	AVector c;

#ifndef ASMMATH_NO_SSE41
	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"dpps $0xFF, %%xmm0, %%xmm0\n\t"
		"sqrtps %%xmm0, %%xmm0\n\t"
		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a)
		);
#else
	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"mulps %%xmm0, %%xmm0\n\t"			// v^2
		"movaps %%xmm0, %%xmm1\n\t"
		"shufps $0xB1, %%xmm1, %%xmm1\n\t"	// 0xB1 = 10110001
		"addps %%xmm0, %%xmm1\n\t"			// xmm1 -> (x, y, z, w)+(y, x, w, z)=(x+y, x+y, z+w, z+w)
		"movaps %%xmm1, %%xmm0\n\t"
		"shufps $0x0A, %%xmm1, %%xmm1\n\t"	// 0x0A = 00001010
		"addps %%xmm0, %%xmm1\n\t"			// xmm1 -> (x+y+z+w, , , )
		"sqrtps %%xmm1, %%xmm1\n\t"
		"movups %%xmm1, %0\n\t"
		: "=m"(c)
		: "r"(&a)
		);
#endif

	return c;
	}


AVector AVectorDot(const AVector& a, const AVector& b)
	{
	AVector c;
#ifndef ASMMATH_NO_SSE41
	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"movups (%2), %%xmm1\n\t"
		"dpps $0xFF, %%xmm1, %%xmm0\n\t"
		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a), "r"(&b)
		);
#else
	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"movups (%2), %%xmm1\n\t"
		"mulps %%xmm1, %%xmm0\n\t"			// xmm0 -> (a1+b1, , , )
		"movaps %%xmm0, %%xmm1\n\t"			// xmm1 = xmm0
		"shufps $0xB1, %%xmm1, %%xmm1\n\t"	// 0xB1 = 10110001
		"addps %%xmm1, %%xmm0\n\t"			// xmm1 -> (x, y, z, w)+(y, x, w, z)=(x+y, x+y, z+w, z+w)
		"movaps %%xmm0, %%xmm1\n\t"			// xmm1 = xmm0
		"shufps $0x0A, %%xmm1, %%xmm1\n\t"	// 0x0A = 00001010
		"addps %%xmm1, %%xmm0\n\t"			// xmm1 -> (x+y+z+w, , , )
		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a), "r"(&b)
		);
#endif

	return c;
	}

AVector AVectorCross(const AVector& a, const AVector& b)
	{
	// return _mm_sub_ps(
    // _mm_mul_ps(_mm_shuffle_ps(a, a, _MM_SHUFFLE(3, 0, 2, 1)), _mm_shuffle_ps(b, b, _MM_SHUFFLE(3, 1, 0, 2))),
    // _mm_mul_ps(_mm_shuffle_ps(a, a, _MM_SHUFFLE(3, 1, 0, 2)), _mm_shuffle_ps(b, b, _MM_SHUFFLE(3, 0, 2, 1)))

	AVector c;

	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"movups (%2), %%xmm1\n\t"

		"movups %%xmm0, %%xmm2\n\t"			// 3102 - D2	// 3021 - C9
		"movups %%xmm1, %%xmm3\n\t"			// 3102 - 87	// 3021 - 63
		"shufps $0xD2, %%xmm2, %%xmm2\n\t"	// xmm2 -> (w, y, x, z)
		"shufps $0xC9, %%xmm3, %%xmm3\n\t"	// xmm3 -> (w, x, z, y)
		"mulps %%xmm3, %%xmm2\n\t"			// xmm2 -> (w*w, y*x, x*z, z*y)

		"shufps $0xC9, %%xmm0, %%xmm0\n\t"	// xmm0 -> (w, x, z, y)
		"shufps $0xD2, %%xmm1, %%xmm1\n\t"	// xmm1 -> (w, y, x, z)
		"mulps %%xmm1, %%xmm0\n\t"			// xmm0 -> (w*w, y*x, x*z, z*y)

		"subps %%xmm2, %%xmm0\n\t"
		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a), "r"(&b)
		);

	return c;
	}


AVector AVectorAbsolute(const AVector& a)
	{
	return AVector(fabs(a.x), fabs(a.y), fabs(a.z), fabs(a.w));
	}

AVector AVectorClamp(const AVector& a, const AVector& min, const AVector& max)
	{
	AVector c;

	asm volatile(
		"movups (%1), %%xmm0\n\t"
		"movups (%2), %%xmm1\n\t"
		"movups (%3), %%xmm2\n\t"

		"minps %%xmm2, %%xmm0\n\t"
		"maxps %%xmm1, %%xmm0\n\t"

		"movups %%xmm0, %0\n\t"
		: "=m"(c)
		: "r"(&a), "r"(&min), "r"(&max)
		);

	return c;
	}

/****************** Ulatwienia, uprzyjemnienia ********************************/
AVector AVectorZero()
	{
	return AVector(0.0f, 0.0f, 0.0f, 0.0f);
	}

AVector AVectorSet(float x, float y, float z, float w)
	{
	return AVector(x, y, z, w);
	}

AVector AVectorReplicate(float val)
	{
	return AVector(val, val, val, val);
	}


AVector AVectorTransform(const AVector& a, const AMatrix& m)
	{
	AVector c;

	c.x=AVectorDot(a, m.row[0]).w;
	c.y=AVectorDot(a, m.row[1]).w;
	c.z=AVectorDot(a, m.row[2]).w;
	c.w=AVectorDot(a, m.row[3]).w;

	return c;
	}

AVector AVectorMirror(const AVector& a, const AVector& mirror, const AVector& forward, const AVector& up)
	{
	AVector n=forward^up;
	AVector mp=n*((mirror-a)|n);

	return a+2.0f*mp;
	}


AMatrix AMatrixTranspose(const AMatrix& m)
	{
	return AMatrix(
		AVector(m.row[0].x, m.row[1].x, m.row[2].x, m.row[3].x),
		AVector(m.row[0].y, m.row[1].y, m.row[2].y, m.row[3].y),
		AVector(m.row[0].z, m.row[1].z, m.row[2].z, m.row[3].z),
		AVector(m.row[0].w, m.row[1].w, m.row[2].w, m.row[3].w)
		);
	}

AMatrix AMatrixAdd(const AMatrix& a, const AMatrix& b)
	{
	AMatrix c;

	c.row[0]=AVectorAdd(a.row[0], b.row[0]);
	c.row[1]=AVectorAdd(a.row[1], b.row[1]);
	c.row[2]=AVectorAdd(a.row[2], b.row[2]);
	c.row[3]=AVectorAdd(a.row[3], b.row[3]);

	return c;
	}

AMatrix AMatrixSubtract(const AMatrix& a, const AMatrix& b)
	{
	AMatrix c;

	c.row[0]=AVectorSubtract(a.row[0], b.row[0]);
	c.row[1]=AVectorSubtract(a.row[1], b.row[1]);
	c.row[2]=AVectorSubtract(a.row[2], b.row[2]);
	c.row[3]=AVectorSubtract(a.row[3], b.row[3]);

	return c;
	}

AMatrix AMatrixMultiply(const AMatrix& a, const AMatrix& b)
	{
	AMatrix c;

	const AMatrix at=AMatrixTranspose(a);

	for(int i=0; i<4; ++i)
		{
		c.row[i].x=AVectorDot(at.row[0], b.row[i]).w;
		c.row[i].y=AVectorDot(at.row[1], b.row[i]).w;
		c.row[i].z=AVectorDot(at.row[2], b.row[i]).w;
		c.row[i].w=AVectorDot(at.row[3], b.row[i]).w;
		}

	return c;
	}

AMatrix AMatrixInverse(const AMatrix& mm)
	{
	// http://stackoverflow.com/questions/1148309/inverting-a-4x4-matrix
	// Sprawdzic -> ftp://download.intel.com/design/PentiumIII/sml/24504301.pdf
	AMatrix mi;

	float *m=(float*)&mm;
	float *inv=(float*)&mi;
	float det;

	inv[ 0]= m[ 5]*m[10]*m[15]-m[ 5]*m[11]*m[14]-m[ 9]*m[ 6]*m[15]+m[9]*m[ 7]*m[14]+m[13]*m[ 6]*m[11]-m[13]*m[ 7]*m[10];
	inv[ 4]=-m[ 4]*m[10]*m[15]+m[ 4]*m[11]*m[14]+m[ 8]*m[ 6]*m[15]-m[8]*m[ 7]*m[14]-m[12]*m[ 6]*m[11]+m[12]*m[ 7]*m[10];
	inv[ 8]= m[ 4]*m[ 9]*m[15]-m[ 4]*m[11]*m[13]-m[ 8]*m[ 5]*m[15]+m[8]*m[ 7]*m[13]+m[12]*m[ 5]*m[11]-m[12]*m[ 7]*m[ 9];
	inv[12]=-m[ 4]*m[ 9]*m[14]+m[ 4]*m[10]*m[13]+m[ 8]*m[ 5]*m[14]-m[8]*m[ 6]*m[13]-m[12]*m[ 5]*m[10]+m[12]*m[ 6]*m[ 9];
	inv[ 1]=-m[ 1]*m[10]*m[15]+m[ 1]*m[11]*m[14]+m[ 9]*m[ 2]*m[15]-m[9]*m[ 3]*m[14]-m[13]*m[ 2]*m[11]+m[13]*m[ 3]*m[10];
	inv[ 5]= m[ 0]*m[10]*m[15]-m[ 0]*m[11]*m[14]-m[ 8]*m[ 2]*m[15]+m[8]*m[ 3]*m[14]+m[12]*m[ 2]*m[11]-m[12]*m[ 3]*m[10];
	inv[ 9]=-m[ 0]*m[ 9]*m[15]+m[ 0]*m[11]*m[13]+m[ 8]*m[ 1]*m[15]-m[8]*m[ 3]*m[13]-m[12]*m[ 1]*m[11]+m[12]*m[ 3]*m[ 9];
	inv[13]= m[ 0]*m[ 9]*m[14]-m[ 0]*m[10]*m[13]-m[ 8]*m[ 1]*m[14]+m[8]*m[ 2]*m[13]+m[12]*m[ 1]*m[10]-m[12]*m[ 2]*m[ 9];
	inv[ 2]= m[ 1]*m[ 6]*m[15]-m[ 1]*m[ 7]*m[14]-m[ 5]*m[ 2]*m[15]+m[5]*m[ 3]*m[14]+m[13]*m[ 2]*m[ 7]-m[13]*m[ 3]*m[ 6];
	inv[ 6]=-m[ 0]*m[ 6]*m[15]+m[ 0]*m[ 7]*m[14]+m[ 4]*m[ 2]*m[15]-m[4]*m[ 3]*m[14]-m[12]*m[ 2]*m[ 7]+m[12]*m[ 3]*m[ 6];
	inv[10]= m[ 0]*m[ 5]*m[15]-m[ 0]*m[ 7]*m[13]-m[ 4]*m[ 1]*m[15]+m[4]*m[ 3]*m[13]+m[12]*m[ 1]*m[ 7]-m[12]*m[ 3]*m[ 5];
	inv[14]=-m[ 0]*m[ 5]*m[14]+m[ 0]*m[ 6]*m[13]+m[ 4]*m[ 1]*m[14]-m[4]*m[ 2]*m[13]-m[12]*m[ 1]*m[ 6]+m[12]*m[ 2]*m[ 5];
	inv[ 3]=-m[ 1]*m[ 6]*m[11]+m[ 1]*m[ 7]*m[10]+m[ 5]*m[ 2]*m[11]-m[5]*m[ 3]*m[10]-m[ 9]*m[ 2]*m[ 7]+m[ 9]*m[ 3]*m[ 6];
	inv[ 7]= m[ 0]*m[ 6]*m[11]-m[ 0]*m[ 7]*m[10]-m[ 4]*m[ 2]*m[11]+m[4]*m[ 3]*m[10]+m[ 8]*m[ 2]*m[ 7]-m[ 8]*m[ 3]*m[ 6];
	inv[11]=-m[ 0]*m[ 5]*m[11]+m[ 0]*m[ 7]*m[ 9]+m[ 4]*m[ 1]*m[11]-m[4]*m[ 3]*m[ 9]-m[ 8]*m[ 1]*m[ 7]+m[ 8]*m[ 3]*m[ 5];
	inv[15]= m[ 0]*m[ 5]*m[10]-m[ 0]*m[ 6]*m[ 9]-m[ 4]*m[ 1]*m[10]+m[4]*m[ 2]*m[ 9]+m[ 8]*m[ 1]*m[ 6]-m[ 8]*m[ 2]*m[ 5];

	det=m[0]*inv[0]+m[1]*inv[4]+m[2]*inv[8]+m[3]*inv[12];

	if(det==0.0f)
		return AMatrixIdentity();

	det=1.0f/det;

	for(int i=0; i<16; i++)
		inv[i]*=det;

	return mi;
	}


AMatrix AMatrixAbsolute(const AMatrix& a)
	{
	return AMatrix(AVectorAbsolute(a.row[0]),
				   AVectorAbsolute(a.row[1]),
				   AVectorAbsolute(a.row[2]),
				   AVectorAbsolute(a.row[3]));
	}

AVector AMatrixDeterminant(const AMatrix& mm)
	{
	// Wycinek kodu obracajacego macierz

	AMatrix mi;

	float *m=(float*)&mm;
	float *inv=(float*)&mi;

	inv[ 0]= m[ 5]*m[10]*m[15]-m[ 5]*m[11]*m[14]-m[ 9]*m[ 6]*m[15]+m[9]*m[ 7]*m[14]+m[13]*m[ 6]*m[11]-m[13]*m[ 7]*m[10];
	inv[ 4]=-m[ 4]*m[10]*m[15]+m[ 4]*m[11]*m[14]+m[ 8]*m[ 6]*m[15]-m[8]*m[ 7]*m[14]-m[12]*m[ 6]*m[11]+m[12]*m[ 7]*m[10];
	inv[ 8]= m[ 4]*m[ 9]*m[15]-m[ 4]*m[11]*m[13]-m[ 8]*m[ 5]*m[15]+m[8]*m[ 7]*m[13]+m[12]*m[ 5]*m[11]-m[12]*m[ 7]*m[ 9];
	inv[12]=-m[ 4]*m[ 9]*m[14]+m[ 4]*m[10]*m[13]+m[ 8]*m[ 5]*m[14]-m[8]*m[ 6]*m[13]-m[12]*m[ 5]*m[10]+m[12]*m[ 6]*m[ 9];

	return AVectorReplicate(m[0]*inv[0]+m[1]*inv[4]+m[2]*inv[8]+m[3]*inv[12]);
	}


AMatrix AMatrixRotX(float ang)
	{
	float s, c;
	sincos(ang*M_PI/180, s, c);

	return AMatrix(
		AVector( 1,  0,  0,  0),
		AVector( 0,  c, -s,  0),
		AVector( 0,  s,  c,  0),
		AVector( 0,  0,  0,  1));
	}

AMatrix AMatrixRotY(float ang)
	{
	float s, c;
	sincos(ang*M_PI/180, s, c);

	return AMatrix(
		AVector( c,  0,  s,  0),
		AVector( 0,  1,  0,  0),
		AVector(-s,  0,  c,  0),
		AVector( 0,  0,  0,  1));
	}

AMatrix AMatrixRotZ(float ang)
	{
	float s, c;
	sincos(ang*M_PI/180, s, c);

	return AMatrix(
		AVector( c, -s,  0,  0),
		AVector( s,  c,  0,  0),
		AVector( 0,  0,  1,  0),
		AVector( 0,  0,  0,  1));
	}

AMatrix AMatrixRotate(float angle, const AVector& dir)
	{
	AMatrix m;

	float s, c;
	sincos(angle*M_PI/180.0f, s, c);

	AVector u=AVectorNormalize(dir);
	float uxy=u.x*u.y*(1-c);
	float uyz=u.y*u.z*(1-c);
	float uxz=u.x*u.z*(1-c);
	float ux2=u.x*u.x*(1-c);
	float uy2=u.y*u.y*(1-c);
	float uz2=u.z*u.z*(1-c);
	float uxsin=u.x*s;
	float uysin=u.y*s;
	float uzsin=u.z*s;

	m.row[0]=AVector(c+ux2, uxy+uzsin, uxz-uysin);
	m.row[1]=AVector(uxy-uzsin, c+uy2, uyz+uxsin);
	m.row[2]=AVector(uxz+uysin, uyz-uxsin, c+uz2);
	m.row[3]=AVector(0, 0, 0, 1);

	return m;
	}

AMatrix AMatrixSkew(const AVector& a)
	{
	return AMatrix(
		AVector(   0, -a.z,  a.y, 0),
		AVector( a.z,    0, -a.x, 0),
		AVector(-a.y,  a.x,    0, 0),
		AVector(   0,    0,    0, 0));
	}

AMatrix AMatrixZero()
	{
	return AMatrix();
	}

AMatrix AMatrixIdentity()
	{
	return AMatrix(AVector(1), AVector(0, 1), AVector(0, 0, 1), AVector(0, 0, 0, 1));
	}

AMatrix AMatrixFill(float val)
	{
	return AMatrix(AVectorReplicate(val), AVectorReplicate(val), AVectorReplicate(val), AVectorReplicate(val));
	}

AMatrix AMatrixDiagonal(float val)
	{
	return AMatrix(AVector(val), AVector(0, val), AVector(0, 0, val), AVector(0, 0, 0, val));
	}

#ifndef UTILS_TEST
#define UTILS_TEST
#endif // lolz
#ifdef UTILS_TEST
#include <cstdlib>
#include <cmath>
#include <ctime>
#include <cstdio>
#include <fstream>

const float DEGTORAD=3.141592653589793f/180.0f;

//int main()
void asmmath_h_main()
	{
	const float EPS=0.000001f;

	AVector a, b, c;

	a=AVector(
		rand()/(RAND_MAX+1.0f),
		rand()/(RAND_MAX+1.0f),
		rand()/(RAND_MAX+1.0f),
		rand()/(RAND_MAX+1.0f));
	b=AVector(
		rand()/(RAND_MAX+1.0f),
		rand()/(RAND_MAX+1.0f),
		rand()/(RAND_MAX+1.0f),
		rand()/(RAND_MAX+1.0f));

	printf("A: %f %f %f %f\n", a.x, a.y, a.z, a.w);
	printf("B: %f %f %f %f\n", b.x, b.y, b.z, b.w);

/**********************************************************/
	//c=AVectorAdd(a, b);
	c=a+b;
	printf("AVectorAdd...");
	if(	fabs(a.x+b.x-c.x)<EPS &&
		fabs(a.y+b.y-c.y)<EPS &&
		fabs(a.z+b.z-c.z)<EPS &&
		fabs(a.w+b.w-c.w)<EPS)
		printf("OK\n");
	else
		printf("FAIL\n");

/**********************************************************/
	//c=AVectorSubtract(a, b);
	c=a-b;
	printf("AVectorSubtract...");
	if(	fabs(a.x-b.x-c.x)<EPS &&
		fabs(a.y-b.y-c.y)<EPS &&
		fabs(a.z-b.z-c.z)<EPS &&
		fabs(a.w-b.w-c.w)<EPS)
		printf("OK\n");
	else
		printf("FAIL\n");

/**********************************************************/
	//c=AVectorMultiply(a, b);
	c=a*b;
	printf("AVectorMultiply...");
	if(	fabs(a.x*b.x-c.x)<EPS &&
		fabs(a.y*b.y-c.y)<EPS &&
		fabs(a.z*b.z-c.z)<EPS &&
		fabs(a.w*b.w-c.w)<EPS)
		printf("OK\n");
	else
		printf("FAIL\n");

/**********************************************************/
	//c=AVectorDivide(a, b);
	c=a/b;
	printf("AVectorDivide...");
	if(	fabs(a.x/b.x-c.x)<EPS &&
		fabs(a.y/b.y-c.y)<EPS &&
		fabs(a.z/b.z-c.z)<EPS &&
		fabs(a.w/b.w-c.w)<EPS)
		printf("OK\n");
	else
		printf("FAIL\n");

	/**********************************************************/
	//a=AVector(1, 0, 0, 0);
	c=AVectorNormalize(a);
	printf("AVectorNormalize...");
	if( fabs(sqrt(c.x*c.x+c.y*c.y+c.z*c.z+c.w*c.w)-1.0f)<EPS )
		printf("OK\n");
	else
		printf("FAIL\n");

	//printf("C: %f %f %f %f\n", c.x, c.y, c.z, c.w);

	/**********************************************************/
	c=AVectorLength(a);
	printf("AVectorLength...");
	if( fabs(sqrt(a.x*a.x+a.y*a.y+a.z*a.z+a.w*a.w)-c.w)<EPS )
		printf("OK\n");
	else
		printf("FAIL\n");

	/**********************************************************/
	c=AVectorDot(a, b);
	//c=a|b;

	printf("AVectorDot...");
	if(	fabs(c.x-(a.x*b.x+a.y*b.y+a.z*b.z+a.w*b.w))<EPS &&
		fabs(c.y-(a.x*b.x+a.y*b.y+a.z*b.z+a.w*b.w))<EPS &&
		fabs(c.z-(a.x*b.x+a.y*b.y+a.z*b.z+a.w*b.w))<EPS &&
		fabs(c.w-(a.x*b.x+a.y*b.y+a.z*b.z+a.w*b.w))<EPS)
		printf("OK\n");
	else
		printf("FAIL\n");

	/**********************************************************/
	a=AVector(1, 0, 0, 0);
	b=AVector(0, 1, 0, 0);
	//c=AVectorCross(a, b);
	c=a^b;

	printf("AVectorCross...");
	if(	fabs(c.x)<EPS &&
		fabs(c.y)<EPS &&
		fabs(c.z-1.0f)<EPS &&
		fabs(c.w)<EPS)
		printf("OK\n");
	else
		printf("FAIL\n");

	/**********************************************************/
	AMatrix m1, m2;
	AMatrix mr;

	m1=AMatrixRotX(180);
	m2=AMatrixRotY(90);

	//mr=AMatrixMultiply(m1, m2);
	mr=m1*m2;

	//for(int i=0; i<4; ++i)
	//	printf("%f %f %f %f\n", mr.row[i].x, mr.row[i].y, mr.row[i].z, mr.row[i].w);

	/**********************************************************/
	/*m1=AMatrixRotZ(90);
	a=AVector(1, 0, 0, 0);

	for(int i=0; i<4; ++i)
		printf("%f %f %f %f\n", m1.row[i].x, m1.row[i].y, m1.row[i].z, m1.row[i].w);

	//c=AVectorTransform(a, m1);
	c=m1*a;
	printf("C: %f %f %f %f\n", c.x, c.y, c.z, c.w);

	//c=AVectorTransform(c, m1);
	c=m1*c;
	printf("C: %f %f %f %f\n", c.x, c.y, c.z, c.w);

	//c=AVectorTransform(c, m1);
	c=m1*c;
	printf("C: %f %f %f %f\n", c.x, c.y, c.z, c.w);*/

	/**********************************************************/
	/*AVector point(1, 1);
	AVector mirror(0, 0);
	AVector forward(1, 0);

	c=AVectorMirror(point, mirror, forward);

	printf("mirror: %f %f %f %f\n", c.x, c.y, c.z, c.w);*/

	/**********************************************************/
	/*m1=AMatrixRotZ(45.0f)*AMatrixRotY(30.0f)*AMatrixRotX(15.0f);

	m2=AMatrixInverse(m1);

	m1=m1*m2;

	for(int i=0; i<4; ++i)
		printf("%f %f %f %f\n", m1.row[i].x, m1.row[i].y, m1.row[i].z, m1.row[i].w);*/

	/**********************************************************/
	/*a=AVector(1, 0, 0);
	b=AVector(0, 1, 0);
	m1=AMatrixSkew(a);
	m1=m1*AMatrixTranspose(m1);

	//c=m1*b;

	//printf("C: %f %f %f %f\n", c.x, c.y, c.z, c.w);
	for(int i=0; i<4; ++i)
		printf("%f %f %f %f\n", m1.row[i].x, m1.row[i].y, m1.row[i].z, m1.row[i].w);*/

	float z;

	a=AVector(3, 4, 0);
	b=AVector(0, 0, 2);
	c=(a^b)^a;
	printf("C: %f %f %f %f\n", c.x, c.y, c.z, c.w);

	c=-a*(b|a)+b*(a|a);
	printf("C: %f %f %f %f\n", c.x, c.y, c.z, c.w);

	c=(a^b)*(a^b);
	printf("C: %f %f %f %f\n", c.x, c.y, c.z, c.w);

	printf("%f\n", AVectorLength(c).w);

	z=((a^b)^a)|b;
	printf("Z: %f\n", z);

	z=(-a*(b|a)|b)+(b*(a|a)|b);
	printf("Z: %f\n", z);

	//return 0;
	}

#endif

