#include "lua.h"

#ifdef WIN32
	#define WINDOWS
#else
	#ifdef WIN64
		#define WINDOWS
	#else
		#define LINUX
	#endif
#endif

extern "C"
	{
	#ifdef WINDOWS
		#include <lua/lua.h>
		#include <lua/lualib.h>
		#include <lua/lauxlib.h>
	#else
		#include <lua.h>
		#include <lualib.h>
		#include <lauxlib.h>
	#endif
	}

#include <sstream>

#include "resource.h"

Lua::Lua(): luaState(NULL), status(0), loaded(0), args(0)
	{
	//init();
	//error();
	}

Lua::~Lua()
	{
	if(luaState)
		lua_close(luaState);
	luaState=NULL;
	}

void Lua::init()
	{
	if(luaState)
		{
		if(luaState)
			lua_close(luaState);
		luaState=NULL;
		}
	status=0;
	loaded=0;
	args=0;
	luaState=luaL_newstate();
	if(luaState==NULL)
		{
		printf("FAIL: Lua.Lua: Nie udalo sie uruchomic lua\n");
		return;
		}
	luaL_openlibs(luaState);
	}

void Lua::error()
	{
	if(status==0)
		return;
	const char *str=lua_tostring(luaState, -1);
	if(str!=NULL)
		printf("FAIL: Lua.update: Error: %s\n", str);
	lua_pop(luaState, -1);
	status=0;
	}

/*int Lua::setArgs()
	{
	int args=intArgs.size()+strArgs.size();
	while(true)
		{
		if(intArgs.size()==0 && strArgs.size()==0)
			break;
		if(intArgs.size()>0 && strArgs.size()>0)
			{
			if(intArgs.front().id<strArgs.front().id)
				{
				lua_pushnumber(luaState, intArgs.front().arg);
				intArgs.pop_front();
				}
			else
				{
				lua_pushstring(luaState, strArgs.front().str.c_str());
				strArgs.pop_front();
				}
			}
		else if(intArgs.size()>0)
			{
			lua_pushnumber(luaState, intArgs.front().arg);
			intArgs.pop_front();
			}
		else
			{
			lua_pushstring(luaState, strArgs.front().str.c_str());
			strArgs.pop_front();
			}
		}
	return args;
	}*/

bool Lua::openFile(const std::string& name)
	{
	if(name.size()<1)
		{
		printf("FAIL: Lua.openFile: Brak nazwy pliku\n");
		return 0;
		}
	//status=luaL_loadfile(luaState, name.c_str());

	char *script=Resource::load(name);
	if(!script)
		{
		printf("FAIL: Lua.openFile: Nie udalo sie wczytac pliku\n");
		return 0;
		}

	status=luaL_loadbuffer(luaState, script, Resource::getLastSize(), "line");

	delete [] script;

	setGlobalStr("__PATH__", name.c_str());

	if(status==0)
		status=lua_pcall(luaState, 0, 0, 0);
	else
		{
		error();
		return 0;
		}
	if(status!=0)
		{
		error();
		return 0;
		}

	loaded=1;
	return 1;
	}

bool Lua::regFunc(Lua::LuaFunction funcPtr, const std::string& name)
	{
	if(funcPtr==NULL || name.size()<1)
		{
		printf("FAIL: Lua.regFunc: NULL\n");
		return 0;
		}
	//lua_pushcfunction(luaState, funcPtr);
	lua_pushcclosure(luaState, funcPtr, 0);
    lua_setglobal(luaState, name.c_str());
    error();
    return 1;
	}

bool Lua::callFunc(const std::string& name)
	{
	if(name.size()<1)									// Jesli ci nie wypisze bledu, to pewnie przesylasz null'a lub nie otwarles pliku o; // Olane sprawdzanie otwarcia, niby po co ma byc~?
		return 0;
	if(status==0)
		{
		lua_getglobal(luaState, name.c_str());
		lua_insert(luaState, lua_gettop(luaState)-args);
		status=lua_pcall(luaState, args, LUA_MULTRET, 0);	// luastate, args, returns
		fflush(stdout);
		//status=lua_pcall(luaState, 0, LUA_MULTRET, 0);
		}

	args=0;

	error();											// Tylko wywolywanie funkcji, ta linia nie oznacza, ze zawsze bedzie tutaj blad~
	return 1;
	}

bool Lua::callLine(const std::string& line)
	{
	if(line.size()<1)
		return 0;
	//int length=0;
	//while(line[length++]);								// Mozna by przerobic na strlen(), ale nie wiem, co bedzie szybsze i lepsze~
	status=luaL_loadbuffer(luaState, line.c_str(), line.size(), "line");
	if(status==0)
		status=lua_pcallk(luaState, args, 0, 0, 0, 0);
	else
		{
		error();
		return 0;
		}

	args=0;

	loaded=1;
	return 1;
	}

bool Lua::callRef(int ref)
	{
	lua_rawgeti(luaState, LUA_REGISTRYINDEX, ref);

	if(!lua_isfunction(luaState, -1))
		{
		lua_settop(luaState, 0);

		return false;
		}

	lua_insert(luaState, lua_gettop(luaState)-args);	// Przesun wskaznik na funkcje na spod stosu

	status=lua_pcall(luaState, args, LUA_MULTRET, 0);

	args=0;

	error();
	return true;
	}

bool Lua::addNumArg(float arg)
	{
	if(!loaded)
		return false;
	//lua_pushnumber(luaState, arg);
	//intArgs.push_back(IntArg(intArgs.size()+strArgs.size(), arg));

	lua_pushnumber(luaState, arg);
	++args;

	return true;
	}

bool Lua::addStrArg(const std::string& str)
	{
	if(!loaded)
		return false;
	//lua_pushstring(luaState, str);
	//strArgs.push_back(StrArg(intArgs.size()+strArgs.size(), str));

	lua_pushstring(luaState, str.c_str());
	++args;

	return true;
	}

/***** Ustawianie wartosci globalnych *****/
int Lua::getGlobalVal(const std::string& name)
	{
	lua_pushstring(luaState, name.c_str());
	lua_gettable(luaState, LUA_REGISTRYINDEX);

	return lua_tonumber(luaState, -1);
	}

const char* Lua::getGlobalStr(const std::string& name)
	{
	lua_pushstring(luaState, name.c_str());
	lua_gettable(luaState, LUA_REGISTRYINDEX);
	return lua_tostring(luaState, -1);
	}

const void* Lua::getGlobalPtr(const std::string& name)
	{
	lua_pushstring(luaState, name.c_str());
	lua_gettable(luaState, LUA_REGISTRYINDEX);
	return lua_topointer(luaState, -1);
	}


void Lua::setGlobalVal(const std::string& name, int val)	// teoretycznie lua przechowuje double
	{
	lua_pushstring(luaState, name.c_str());
	lua_pushinteger(luaState, val);
	lua_settable(luaState, LUA_REGISTRYINDEX);
	}

void Lua::setGlobalStr(const std::string& name, const std::string& str)
	{
	lua_pushstring(luaState, name.c_str());
	lua_pushstring(luaState, str.c_str());
	lua_settable(luaState, LUA_REGISTRYINDEX);
	}

void Lua::setGlobalPtr(const std::string& name, void *ptr)
	{
	lua_pushstring(luaState, name.c_str());
	lua_pushlightuserdata(luaState, ptr);
	lua_settable(luaState, LUA_REGISTRYINDEX);
	}

int Lua::setGlobalRefFunction(const std::string& fctName)
	{
	lua_getglobal(luaState, fctName.c_str());

	if(!lua_isfunction(luaState, -1))
		{
		lua_settop(luaState, 0);

		return LUA_NOREF;
		}

	return luaL_ref(luaState, LUA_REGISTRYINDEX);
	}

bool Lua::functionExists(const std::string& name)
	{
	bool ret;

	lua_getglobal(luaState, name.c_str());

	ret=lua_isfunction(luaState, -1);

	lua_settop(luaState, 0);

	return ret;
	}

void Lua::createFunction(const std::string& name)
	{
	std::stringstream ss;
	ss << "function " << name << "() end";

	callLine(ss.str());
	}

/*static int lKillApp(lua_State *L)						// <- tego typu ma byc kazda funkcja ktora chcemy dodac do lua	// nazwa ofc dowolna
	{
	int tmp=lua_tonumber(luaState, 1);	// Pobranie argumentow
	lua_pushnumber(L, ++tmp);			// 'Return'
	return 2;							// Ilosc 'returnow'
	}*/
