#pragma once

#include <list>
#include <string>

struct lua_State;

class Lua
	{
	public:
		typedef int (*LuaFunction)(lua_State *);

	private:
		struct IntArg
			{
			IntArg(int id, int arg): id(id), arg(arg) {};

			int id;
			int arg;
			};
		struct StrArg
			{
			StrArg(int id, const std::string& str): id(id), str(str) {};

			int id;
			std::string str;
			};

		lua_State *luaState;
		int status;
		bool loaded;
		//std::list<IntArg> intArgs;
		//std::list<StrArg> strArgs;
		int args;

		void error();
		//int setArgs();
		//static void init();		// Fajna funkcja w ktorej - w zamierzeniu - bedzie rejestracja wszystkich funkcji
	public:
		Lua();
		~Lua();

		void init();				// Zamyka jesli byl otwarty i otwiera nowego luaState
	// Te funkcje niby sa boolowe, ale polowa z nich zawsze zwroci 1
		bool openFile(const std::string& name);									// Otwiera (i wywoluje raz) plik fileName
		bool regFunc(LuaFunction fptr, const std::string& name);	// Rejestracja funkcji funcName w lua

		bool callFunc(const std::string& name);									// Wywolanie funkcji funcName (chyba nie przejdzie wywolanie recznie dodanej funkcji)
		bool callLine(const std::string& line);									// Wywolanie podanego kodu
		bool callRef(int ref);
		bool addIntArg(int arg) {return addNumArg((float)arg);}
		bool addNumArg(float arg);
		bool addStrArg(const std::string& str);

		inline bool operator() (const std::string& name) {return callFunc(name);}
		inline bool operator= (const std::string& line) {return callLine(line);}
		inline operator bool () {return isOk()&isLoaded();}

		inline bool isOk() {return (status==0);}
		inline bool isLoaded() {return loaded;}
		inline lua_State* getState() const {return luaState;}

		int getGlobalVal(const std::string& valName);
		const char* getGlobalStr(const std::string& strName);
		const void* getGlobalPtr(const std::string& ptrName);

		void setGlobalVal(const std::string& valName, int val);
		void setGlobalStr(const std::string& strName, const std::string& str);
		void setGlobalPtr(const std::string& ptrName, void *ptr);

		int setGlobalRefFunction(const std::string& fctName);
		bool functionExists(const std::string& name);
		void createFunction(const std::string& name);

		/**** Nowy zestaw funkcji ****/
		/*void init();
		void clear();

		bool open(const std::string& name);
		bool line(const std::string& line);
		bool call(const std::string& name);

		inline bool operator() (const std::string& name) {return call(name);}
		inline bool operator= (const std::string& line) {return line(line);}
		inline operator bool () {return isOk()&isLoaded();}

		inline bool isOk() {return (status==0);}
		inline bool isLoaded() {return loaded;}

		void pushArgInt(int val);
		void pushArgStr(const std::string& str);
		void popArg();
		//void popInt();
		//void popStr();

		void addTab(const std::string& path);
		void addVal(const std::string& name, int val);
		void addStr(const std::string& name, const std::string& str);
		void addPtr(const std::string& name, void *ptr);
		void addFunc(const std::string& name, LuaFunction fptr);

		bool isFunc(const std::string& name);
		bool isVar(const std::string& name);*/
	};

/*
** Otwieramy plik openFile, to co luzem zostaje wykonane
** Wywolujemy funkcje callFunc, jesli funkcja pojawila sie w pliku otwartym przy pomocy openFile lub callLine (! o; ) - zostanie wykonana
** Wywolanie callLine nie kasuje tego, co pojawilo sie po wykonaniu pliku, callFunc lub callLine
** Tak samo zreszta jak wywolanie callFunc~
*/
